"""
Copyright 2015-2023 National Technology & Engineering Solutions of Sandia, LLC (NTESS).
Under the terms of Contract DE-NA0003525 with NTESS, the U.S. Government retains certain rights in this software.

You should have received a copy of the GNU General Public License along with HyRAM+.
If not, see https://www.gnu.org/licenses/.
"""

import os
import unittest
import json

import numpy as np
from scipy.constants import milli, kilo, liter, bar, pi, minute

from hyram.phys import Fluid, Orifice, Jet
from hyram.utilities import misc_utils

import utils

"""
NOTE: if running from IDE like pycharm, make sure cwd is hyram/ and not hyram/tests.
"""

# Flags to enable command line output, pyplots, and text file output
VERBOSE = False
CREATE_PLOTS = False
CREATE_OUTPUT = False

# Absolute paths to input/output data
DATA_LOC = os.path.join(os.path.dirname(__file__), 'data')
OUTPUT_LOC = os.path.join(misc_utils.get_temp_folder(), 'validation-jetplume')
OUTPUT_FILE = os.path.join(OUTPUT_LOC, 'Validation.txt')
LIMITS_FILE = os.path.join(DATA_LOC, 'Limits-jetplume.json')


class Test_Molkov_2012(unittest.TestCase):
    """
    Tests for table 5-3 from Molkov
    """
    def setUp(self):
        csv_data = utils.read_csv(os.path.join(DATA_LOC, 'Molkov2012BookData.csv'), header=1)

        with open(LIMITS_FILE) as f:
            self.limits = json.load(f)[__class__.__name__]

        # Set parameters from data
        self.jets = []
        self.data = dict()
        air = Fluid(T = 298, P = 101325, species = 'air')
        for idx in range(0, len(csv_data['T']), 2):
            try:
                gas = Fluid(species='h2', T=csv_data['T'][idx], P=csv_data['P'][idx]*1e6)
                orifice = Orifice(d=csv_data['D'][idx]*1e-3)
                self.jets.append(Jet(gas, orifice, air, lam=1.16, theta0=pi/2))

                # Datapoints come in pairs, add to self.data dict
                for key in csv_data:
                    if key in self.data:
                        self.data[key].append(csv_data[key][idx])
                        self.data[key].append(csv_data[key][idx+1])

                    else:
                        self.data[key] = list()
                        self.data[key].append(csv_data[key][idx])
                        self.data[key].append(csv_data[key][idx+1])
            except:
                pass

    def test_concentration(self):
        trial_num = 0
        prev_title = ''
        for idx in range(0, len(self.data['T']), 2):
            jet_idx = int(idx/2)
            x1 = self.data['x'][idx] / (self.data['D'][idx] * np.sqrt(self.data['rho_n'][idx] / self.jets[jet_idx].ambient.rho * 1e-6))
            x2 = self.data['x'][idx+1] / (self.data['D'][idx+1] * np.sqrt(self.data['rho_n'][idx+1] / self.jets[jet_idx].ambient.rho * 1e-6))
            exp_x = [x1, x2]

            c1 = self.data['Y'][idx]
            c2 = self.data['Y'][idx+1]
            exp_concentration = [c1, c2]

            calc_x = self.jets[jet_idx].S / (self.data['D'][idx] * 1e-3 * np.sqrt(self.jets[jet_idx].developing_flow.fluid_orifice.rho / self.jets[jet_idx].ambient.rho))
            calc_concentration = self.jets[jet_idx].Y_cl

            final_data = utils.interp_and_trim_data(calculated_x=calc_x, calculated_y=calc_concentration,
                                                    experiment_x=exp_x, experiment_y=exp_concentration)

            title = self.data['Author-Year'][idx]
            if title != prev_title:
                prev_title = title
                trial_num = 1
            else:
                trial_num += 1

            title = "Molkov 2012, Table 5.3 - " + str(title) + ', Sample ' + str(trial_num)
            error_limits = self.limits[title]
            error = utils.get_error(exp_vals=final_data['exp_y'],
                                    calc_vals=final_data['calc_y'],
                                    error_limits=error_limits,
                                    units='',
                                    msg=title,
                                    output_filename=OUTPUT_FILE,
                                    create_output=CREATE_OUTPUT,
                                    verbose=VERBOSE)

            with self.subTest():
                self.assertLessEqual(error['Max Absolute Error'], error_limits['Max Absolute Error'], f"{title}: Maximum Absolute Error out of range")
            with self.subTest():
                self.assertLessEqual(error['Avg Absolute Error'], error_limits['Avg Absolute Error'], f"{title}: Average Absolute Error out of range")
            with self.subTest():
                self.assertLessEqual(error['Max Percent Error'], error_limits['Max Percent Error'], f"{title}: Maximum Percent Error out of range")
            with self.subTest():
                self.assertLessEqual(error['Avg Percent Error'], error_limits['Avg Percent Error'], f"{title}: Average Percent Error out of range")
            with self.subTest():
                self.assertGreaterEqual(error['R2'], error_limits['R2'], f"{title}: R Squared Value out of range")

            if CREATE_PLOTS:
                utils.create_plots(output_dir=OUTPUT_LOC,
                                x=final_data['calc_x'],
                                exp_y=final_data['exp_y'],
                                calc_y=final_data['calc_y'],
                                max_error=error['Max Absolute Error'],
                                avg_error=error['Avg Absolute Error'],
                                smape=error['SMAPE'],
                                title=title,
                                xlabel=r'Normalized Distance ($\frac{x}{d\sqrt{\rho_{\rm throat}/\rho_{\rm amb}}}$)',
                                ylabel='Centerline Mass Fraction')


class Test_HoufSchefer_2008_fig5(unittest.TestCase):
    """
    Tests for mass fraction curve at various distances, data from Houf and Schefer 2008 Figure 5
    """

    def setUp(self):
        data = utils.read_csv(os.path.join(DATA_LOC, 'houfschefer-2008-fig5.csv'), header=1)

        with open(LIMITS_FILE) as f:
            self.limits = json.load(f)[__class__.__name__]

        # Set parameters
        diameter = 1.905 * milli #m
        density = 0.0838 #kg/m^3
        flowrate = 22.9 * liter/minute
        amb_temp = 294 #K
        amb_pressure = 100 * kilo #Pa
        axial_dists = [10, 25, 50, 75, 100] #Z/D, axial distances from the flame

        air = Fluid(T=amb_temp, P=amb_pressure, species='air')
        gas = Fluid(T=amb_temp, P=amb_pressure, species='H2')
        orifice = Orifice(d=diameter)
        jet = Jet(gas, orifice, air, mdot=density*flowrate, theta0=pi/2)

        # Build dicts of calculated and experimental values
        self.output_data = []
        for zdist in axial_dists:
            profile = jet._radial_profile(zdist * orifice.d, ind_var='Y', nB=20)

            distance = profile[0] * 1000
            mass_fraction = profile[1]

            exp_distance = data[f'z{zdist}x']
            exp_mass_fraction = data[f'z{zdist}y']

            # Trim to [-20, 20]
            final_data = utils.interp_and_trim_data(calculated_x=distance, calculated_y=mass_fraction,
                                                    experiment_x=exp_distance, experiment_y=exp_mass_fraction,
                                                    lower_bound=-20, upper_bound=20)

            entry = [final_data['calc_x'], final_data['calc_y'], final_data['exp_y']]
            self.output_data.append(entry)

    def test_zdists(self):
        # Note: 'zdist' is the normalized axial distance Z/D, no units
        zdists = ['10', '25', '50', '75', '100']

        for idx, data in enumerate(self.output_data):
            zdist = zdists[idx]
            title = "Houf and Schefer 2008 - Figure 5, Mass Fraction over Radial Distance, zdist = " + zdist
            calc_distance = data[0]
            calc_mass_fraction = data[1]
            exp_mass_fraction = data[2]

            error_limits = self.limits[title]
            error = utils.get_error(exp_vals=exp_mass_fraction,
                                    calc_vals=calc_mass_fraction,
                                    error_limits=error_limits,
                                    units='',
                                    msg=title,
                                    output_filename=OUTPUT_FILE,
                                    create_output=CREATE_OUTPUT,
                                    verbose=VERBOSE)

            with self.subTest():
                self.assertLessEqual(error['Max Absolute Error'], error_limits['Max Absolute Error'], f"{title}: Maximum Absolute Error out of range")
            with self.subTest():
                self.assertLessEqual(error['Avg Absolute Error'], error_limits['Avg Absolute Error'], f"{title}: Average Absolute Error out of range")
            with self.subTest():
                self.assertLessEqual(error['Max Percent Error'], error_limits['Max Percent Error'], f"{title}: Maximum Percent Error out of range")
            with self.subTest():
                self.assertLessEqual(error['Avg Percent Error'], error_limits['Avg Percent Error'], f"{title}: Average Percent Error out of range")
            with self.subTest():
                self.assertGreaterEqual(error['R2'], error_limits['R2'], f"{title}: R Squared Value out of range")

            if CREATE_PLOTS:
                utils.create_plots(output_dir=OUTPUT_LOC,
                                x=calc_distance,
                                exp_y=exp_mass_fraction,
                                calc_y=calc_mass_fraction,
                                max_error=error['Max Absolute Error'],
                                avg_error=error['Avg Absolute Error'],
                                smape=error['SMAPE'],
                                title=title,
                                xlabel='Radial Distance (mm)',
                                ylabel='Mass Fraction')


class Test_HoufSchefer_2008_fig8(unittest.TestCase):
    """
    Single test for inverse centerline mole fraction curve over axial distance, data from Houf and Schefer 2008 Figure 8
    """

    def setUp(self):
        data = utils.read_csv(os.path.join(DATA_LOC, 'houfschefer-2008-fig8.csv'), header=1)

        with open(LIMITS_FILE) as f:
            self.limits = json.load(f)[__class__.__name__]

        # Set parameters
        diameter = 1.905 * milli #m
        density = 0.0838 #kg/m^3
        flowrate = 22.9 * liter/minute
        amb_temp = 294 #K
        amb_pressure = 100 * kilo #Pa

        air = Fluid(T=amb_temp, P=amb_pressure, species='air')
        gas = Fluid(T=amb_temp, P=amb_pressure, species='H2')
        orifice = Orifice(d=diameter)
        jet = Jet(gas, orifice, air, mdot=density*flowrate, theta0=pi/2)

        # Calculated values
        calc_axial_distance = jet.S / orifice.d
        calc_inverse_mole_fraction = 1 / jet.X_cl

        # Experimental values
        exp_axial_distance = data['dp_x']
        exp_inverse_mole_fraction = data['dp_y']

        # Trim to [0, 80]
        self.final_data = utils.interp_and_trim_data(calculated_x=calc_axial_distance, calculated_y=calc_inverse_mole_fraction,
                                                     experiment_x=exp_axial_distance, experiment_y=exp_inverse_mole_fraction,
                                                     lower_bound=0, upper_bound=80)

    def test_inverse_mole_fraction(self):
        title = "Houf and Schefer 2008 - Figure 8, Inverse Mole Fraction over Axial Distance"
        error_limits = self.limits[title]
        error = utils.get_error(exp_vals=self.final_data['exp_y'],
                                calc_vals=self.final_data['calc_y'],
                                error_limits=error_limits,
                                units='',
                                msg=title,
                                output_filename=OUTPUT_FILE,
                                create_output=CREATE_OUTPUT,
                                verbose=VERBOSE)

        with self.subTest():
            self.assertLessEqual(error['Max Absolute Error'], error_limits['Max Absolute Error'], f"{title}: Maximum Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Absolute Error'], error_limits['Avg Absolute Error'], f"{title}: Average Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Max Percent Error'], error_limits['Max Percent Error'], f"{title}: Maximum Percent Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Percent Error'], error_limits['Avg Percent Error'], f"{title}: Average Percent Error out of range")
        with self.subTest():
            self.assertGreaterEqual(error['R2'], error_limits['R2'], f"{title}: R Squared Value out of range")

        if CREATE_PLOTS:
            utils.create_plots(output_dir=OUTPUT_LOC,
                               x=self.final_data['calc_x'],
                               exp_y=self.final_data['exp_y'],
                               calc_y=self.final_data['calc_y'],
                               max_error=error['Max Absolute Error'],
                               avg_error=error['Avg Absolute Error'],
                               smape=error['SMAPE'],
                               title=title,
                               xlabel='Radial Distance (mm)',
                               ylabel='Mass Fraction')


class Test_HoufSchefer_2008_fig9(unittest.TestCase):
    """
    Tests for inverse mass fraction curve at various axial distances, data from Houf and Schefer 2008 Figure 9
    """
    def setUp(self):
        data = utils.read_csv(os.path.join(DATA_LOC, 'houfschefer-2008-fig9.csv'), header=1)

        with open(LIMITS_FILE) as f:
            self.limits = json.load(f)[__class__.__name__]

        # Set parameters
        diameter = 1.905 * milli #m
        density = 0.0838 #kg/m^3
        flowrate = 22.9 * liter/minute
        amb_temp = 294 #K
        amb_pressure = 100 * kilo #Pa

        air = Fluid(T=amb_temp, P=amb_pressure, species='air')
        gas = Fluid(T=amb_temp, P=amb_pressure, species='H2')
        orifice = Orifice(d=diameter)

        flowrates = np.array([8.497, 13.08, 22.9]) * liter/minute # volumetric flow rates for Fr = 41, 99, 152, 268
        colors = ['blu', 'red', 'bla']

        self.output_data = []
        for idx, flowrate in enumerate(flowrates):
            jet = Jet(gas, orifice, air, mdot=density*flowrate, theta0=pi/2)
            calc_axial_distance = jet.S / orifice.d
            calc_mole_fraction = 1 / jet.X_cl

            exp_axial_distance = data[f'{colors[idx]}_x']
            exp_mole_fraction = data[f'{colors[idx]}_y']

            trimmed_data = utils.interp_and_trim_data(calculated_x=calc_axial_distance, calculated_y=calc_mole_fraction,
                                                    experiment_x=exp_axial_distance, experiment_y=exp_mole_fraction, upper_bound=100)

            self.output_data.append(trimmed_data)

    def test_inverse_mole_fraction(self):
        froud_nums = ['99', '152', '268']
        for idx, data in enumerate(self.output_data):
            title = "Houf and Schefer 2008 - Figure 9, Inverse Mass Fraction over Axial Distance, Fr = " + froud_nums[idx]

            error_limits = self.limits[title]
            error = utils.get_error(exp_vals=data['exp_y'],
                                    calc_vals=data['calc_y'],
                                    error_limits=error_limits,
                                    units='',
                                    msg=title,
                                    output_filename=OUTPUT_FILE,
                                    create_output=CREATE_OUTPUT,
                                    verbose=VERBOSE)

            with self.subTest():
                self.assertLessEqual(error['Max Absolute Error'], error_limits['Max Absolute Error'], f"{title}: Maximum Absolute Error out of range")
            with self.subTest():
                self.assertLessEqual(error['Avg Absolute Error'], error_limits['Avg Absolute Error'], f"{title}: Average Absolute Error out of range")
            with self.subTest():
                self.assertLessEqual(error['Max Percent Error'], error_limits['Max Percent Error'], f"{title}: Maximum Percent Error out of range")
            with self.subTest():
                self.assertLessEqual(error['Avg Percent Error'], error_limits['Avg Percent Error'], f"{title}: Average Percent Error out of range")
            with self.subTest():
                self.assertGreaterEqual(error['R2'], error_limits['R2'], f"{title}: R Squared Value out of range")

            if CREATE_PLOTS:
                utils.create_plots(output_dir=OUTPUT_LOC,
                                   x=data['calc_x'],
                                   exp_y=data['exp_y'],
                                   calc_y=data['calc_y'],
                                   max_error=error['Max Absolute Error'],
                                   avg_error=error['Avg Absolute Error'],
                                   smape=error['SMAPE'],
                                   title=title,
                                   xlabel='Normalized Axial Distance (z/D)',
                                   ylabel='Inverse Centerline Mole Fraction')


class Test_RugglesEkoto_2012(unittest.TestCase):
    """
    Tests for inverse mass fraction curve and jet (half) width at various axial distances, data from Ruggles and Ekoto 2012
    """
    def setUp(self):
        data = utils.read_csv(os.path.join(DATA_LOC, 'rugglesekoto-2012-fig7.csv'), header=2)

        with open(LIMITS_FILE) as f:
            self.limits = json.load(f)[__class__.__name__]

        # Set parameters
        diameter = 0.75*2 *milli#meters
        temp = 295.4 #K
        pressure = 983.2 *kilo#Pa
        discharge_coeff = 0.979
        amb_temp = 296 #K
        amb_pressure = 98.37 *kilo#Pa

        air = Fluid(T=amb_temp, P=amb_pressure, species='air')
        gas = Fluid(T=temp, P=pressure, species='h2')
        orifice = Orifice(d=diameter, Cd=discharge_coeff)
        jet = Jet(gas, orifice, air, theta0=pi/2)

        calc_axial_distance = jet.y / (orifice.d / 2)
        calc_mass_fraction = 1 / jet.Y_cl
        calc_jet_width = jet.B * jet.lam * np.sqrt(-np.log(.5)) / milli

        exp_axial_distance_1 = data['x1']
        exp_mass_fraction = data['y1']

        exp_axial_distance_2 = data['x3']
        exp_jet_width = data['y3']

        self.mass_fraction_data = utils.interp_and_trim_data(calculated_x=calc_axial_distance, calculated_y=calc_mass_fraction,
                                                            experiment_x=exp_axial_distance_1, experiment_y=exp_mass_fraction,
                                                            upper_bound=300)

        self.jet_width_data = utils.interp_and_trim_data(calculated_x=calc_axial_distance, calculated_y=calc_jet_width,
                                                            experiment_x=exp_axial_distance_2, experiment_y=exp_jet_width,
                                                            upper_bound=300)

    def test_inverse_mass_fraction(self):
        title = "Ruggles and Ekoto 2012 - Figure 7, Inverse Mass Fraction over Axial Distance"

        error_limits = self.limits[title]
        error = utils.get_error(exp_vals=self.mass_fraction_data['exp_y'],
                                calc_vals=self.mass_fraction_data['calc_y'],
                                error_limits=error_limits,
                                units='',
                                msg=title,
                                output_filename=OUTPUT_FILE,
                                create_output=CREATE_OUTPUT,
                                verbose=VERBOSE)

        with self.subTest():
            self.assertLessEqual(error['Max Absolute Error'], error_limits['Max Absolute Error'], f"{title}: Maximum Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Absolute Error'], error_limits['Avg Absolute Error'], f"{title}: Average Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Max Percent Error'], error_limits['Max Percent Error'], f"{title}: Maximum Percent Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Percent Error'], error_limits['Avg Percent Error'], f"{title}: Average Percent Error out of range")
        with self.subTest():
            self.assertGreaterEqual(error['R2'], error_limits['R2'], f"{title}: R Squared Value out of range")

        if CREATE_PLOTS:
            utils.create_plots(output_dir=OUTPUT_LOC,
                            x=self.mass_fraction_data['calc_x'],
                            exp_y=self.mass_fraction_data['exp_y'],
                            calc_y=self.mass_fraction_data['calc_y'],
                            max_error=error['Max Absolute Error'],
                            avg_error=error['Avg Absolute Error'],
                            smape=error['SMAPE'],
                            title=title,
                            xlabel='Normalized Axial Distance (z/r)',
                            ylabel='Inverse Centerline Mass Fraction')

    def test_jet_width(self):
        title = "Ruggles and Ekoto 2012 - Figure 7, Jet Width over Axial Distance"

        error_limits = self.limits[title]
        error = utils.get_error(exp_vals=self.jet_width_data['exp_y'],
                                calc_vals=self.jet_width_data['calc_y'],
                                error_limits=error_limits,
                                units='mm',
                                msg=title,
                                output_filename=OUTPUT_FILE,
                                create_output=CREATE_OUTPUT,
                                verbose=VERBOSE)

        with self.subTest():
            self.assertLessEqual(error['Max Absolute Error'], error_limits['Max Absolute Error'], f"{title}: Maximum Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Absolute Error'], error_limits['Avg Absolute Error'], f"{title}: Average Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Max Percent Error'], error_limits['Max Percent Error'], f"{title}: Maximum Percent Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Percent Error'], error_limits['Avg Percent Error'], f"{title}: Average Percent Error out of range")
        with self.subTest():
            self.assertGreaterEqual(error['R2'], error_limits['R2'], f"{title}: R Squared Value out of range")

        if CREATE_PLOTS:
            utils.create_plots(output_dir=OUTPUT_LOC,
                            x=self.jet_width_data['calc_x'],
                            exp_y=self.jet_width_data['exp_y'],
                            calc_y=self.jet_width_data['calc_y'],
                            max_error=error['Max Absolute Error'],
                            avg_error=error['Avg Absolute Error'],
                            smape=error['SMAPE'],
                            title=title,
                            xlabel='Normalized Axial Distance (z/r)',
                            ylabel='Jet Half Width (mm)')


class Test_Han_2013_fig3(unittest.TestCase):
    """
    Tests for centerline concentration over distance at various pressures, data from Han et al. 2013
    """
    def setUp(self):
        data_a = utils.read_csv(os.path.join(DATA_LOC, 'han-2013-fig3a.csv'), header=2)
        data_b = utils.read_csv(os.path.join(DATA_LOC, 'han-2013-fig3b.csv'), header=2)
        data_c = utils.read_csv(os.path.join(DATA_LOC, 'han-2013-fig3c.csv'), header=2)
        data_d = utils.read_csv(os.path.join(DATA_LOC, 'han-2013-fig3d.csv'), header=2)

        csv_data = [data_a, data_b, data_c, data_d]

        with open(LIMITS_FILE) as f:
            self.limits = json.load(f)[__class__.__name__]

        # Set parameters
        temp = 293 #K
        amb_temp = 293 #K
        amb_pressure = 101325 #Pa

        air = Fluid(T=amb_temp, P=amb_pressure, species='air')

        self.diameters = np.array([0.5, 0.7, 1]) * milli #m
        self.pressures = np.arange(100, 401, 100) * bar

        self.output_data = dict()
        for file, pressure in enumerate(self.pressures):
            gas = Fluid(T=temp, P=pressure, species='h2')
            self.output_data[pressure] = dict()
            for idx, diameter in enumerate(self.diameters):
                orifice = Orifice(d=diameter)
                jet = Jet(gas, orifice, air, theta0=0, Ymin=1e-4)

                calc_distance = jet.x
                calc_concentration = jet.X_cl

                # Data is in x1/y1, x3/y3, x5/y5
                exp_distance = csv_data[file][f'x{idx*2+1}']
                exp_concentration = csv_data[file][f'y{idx*2+1}']

                self.output_data[pressure][diameter] = utils.interp_and_trim_data(calculated_x=calc_distance, calculated_y=calc_concentration,
                                                            experiment_x=exp_distance, experiment_y=exp_concentration,
                                                            lower_bound=0.5, upper_bound=10)

    def test_centerline_concentration(self):
        pressure_list = ['100', '200', '300', '400']
        diameter_list = ['0.5', '0.7', '1.0']
        for p, pressure in enumerate(self.pressures):
            for d, diameter in enumerate(self.diameters):
                title = "Han et al. 2013 - Figure 3, Centerline Mass Fraction vs. Distance, P0 = " + pressure_list[p] + " bar, d = " + diameter_list[d] + "mm"

                data = self.output_data[pressure][diameter]
                error_limits = self.limits[title]
                error = utils.get_error(exp_vals=data['exp_y'],
                                        calc_vals=data['calc_y'],
                                        error_limits=error_limits,
                                        units='',
                                        msg=title,
                                        output_filename=OUTPUT_FILE,
                                        create_output=CREATE_OUTPUT,
                                        verbose=VERBOSE)

                with self.subTest():
                    self.assertLessEqual(error['Max Absolute Error'], error_limits['Max Absolute Error'], f"{title}: Maximum Absolute Error out of range")
                with self.subTest():
                    self.assertLessEqual(error['Avg Absolute Error'], error_limits['Avg Absolute Error'], f"{title}: Average Absolute Error out of range")
                with self.subTest():
                    self.assertLessEqual(error['Max Percent Error'], error_limits['Max Percent Error'], f"{title}: Maximum Percent Error out of range")
                with self.subTest():
                    self.assertLessEqual(error['Avg Percent Error'], error_limits['Avg Percent Error'], f"{title}: Average Percent Error out of range")
                with self.subTest():
                    self.assertGreaterEqual(error['R2'], error_limits['R2'], f"{title}: R Squared Value out of range")

                if CREATE_PLOTS:
                    utils.create_plots(output_dir=OUTPUT_LOC,
                                    x=data['calc_x'],
                                    exp_y=data['exp_y'],
                                    calc_y=data['calc_y'],
                                    max_error=error['Max Absolute Error'],
                                    avg_error=error['Avg Absolute Error'],
                                    smape=error['SMAPE'],
                                    title=title,
                                    xlabel='Normalized Axial Distance (z/r)',
                                    ylabel='Inverse Centerline Mass Fraction')


class Test_Han_2013_fig6(unittest.TestCase):
    """
    Tests for dilution length over pressure with various orifice diameters, data from Han et al. 2013
    """
    def setUp(self):
        data_a = utils.read_csv(os.path.join(DATA_LOC, 'han-2013-fig6a.csv'), header=2)
        data_b = utils.read_csv(os.path.join(DATA_LOC, 'han-2013-fig6b.csv'), header=2)
        data_c = utils.read_csv(os.path.join(DATA_LOC, 'han-2013-fig6c.csv'), header=2)

        csv_data = [data_a, data_b, data_c]

        with open(LIMITS_FILE) as f:
            self.limits = json.load(f)[__class__.__name__]

        # Set parameters
        temp = 293 #K
        amb_temp = 293 #K
        amb_pressure = 101325 #Pa

        air = Fluid(T=amb_temp, P=amb_pressure, species='air')

        self.diameters = np.array([0.5, 0.7, 1]) * milli #m
        pressures = np.arange(100, 401, 100) * bar

        # Build 2-level dict of jets based on pressure and orifice diameter
        jets = dict()
        for file, pressure in enumerate(pressures):
            gas = Fluid(T=temp, P=pressure, species='h2')
            jets[pressure] = dict()
            for idx, diameter in enumerate(self.diameters):
                orifice = Orifice(d=diameter)
                jets[pressure][diameter] = Jet(gas, orifice, air, theta0=0, Ymin=1e-4)

        lean_flamm_limit = 0.04
        self.lfl_fractions = [lean_flamm_limit, lean_flamm_limit/2, lean_flamm_limit/4]

        self.output_data = dict()
        for file, diameter in enumerate(self.diameters):
            self.output_data[diameter] = dict()
            for idx, fraction in enumerate(self.lfl_fractions):
                calc_pressure = pressures / bar
                calc_dilution_length = [np.interp(fraction, jets[pressure][diameter].X_cl[::-1], jets[pressure][diameter].S[::-1]) for pressure in pressures]

                # Data is in x2/y2, x4/y4, x6/y6
                exp_pressure = csv_data[file][f'x{(idx+1)*2}']
                exp_dilution_length = csv_data[file][f'y{(idx+1)*2}']

                self.output_data[diameter][fraction] = utils.interp_and_trim_data(calculated_x=calc_pressure, calculated_y=calc_dilution_length,
                                                                                  experiment_x=exp_pressure, experiment_y=exp_dilution_length)

    def test_dilution_length(self):
        fractions = ['', '_2', '_4']
        for diameter in self.diameters:
            for f, fraction in enumerate(self.lfl_fractions):
                title = "Han et al. 2013 - Figure 6, Dilution Length vs. Pressure, d = " + str(diameter / milli) + "mm, LFL" + fractions[f]

                data = self.output_data[diameter][fraction]
                error_limits = self.limits[title]
                error = utils.get_error(exp_vals=data['exp_y'],
                                        calc_vals=data['calc_y'],
                                        error_limits=error_limits,
                                        units='m',
                                        msg=title,
                                        output_filename=OUTPUT_FILE,
                                        create_output=CREATE_OUTPUT,
                                        verbose=VERBOSE)

                with self.subTest():
                    self.assertLessEqual(error['Max Absolute Error'], error_limits['Max Absolute Error'], f"{title}: Maximum Absolute Error out of range")
                with self.subTest():
                    self.assertLessEqual(error['Avg Absolute Error'], error_limits['Avg Absolute Error'], f"{title}: Average Absolute Error out of range")
                with self.subTest():
                    self.assertLessEqual(error['Max Percent Error'], error_limits['Max Percent Error'], f"{title}: Maximum Percent Error out of range")
                with self.subTest():
                    self.assertLessEqual(error['Avg Percent Error'], error_limits['Avg Percent Error'], f"{title}: Average Percent Error out of range")
                with self.subTest():
                    self.assertGreaterEqual(error['R2'], error_limits['R2'], f"{title}: R Squared Value out of range")

                if CREATE_PLOTS:
                    utils.create_plots(output_dir=OUTPUT_LOC,
                                    x=data['calc_x'],
                                    exp_y=data['exp_y'],
                                    calc_y=data['calc_y'],
                                    max_error=error['Max Absolute Error'],
                                    avg_error=error['Avg Absolute Error'],
                                    smape=error['SMAPE'],
                                    title=title,
                                    xlabel='Pressure (bar)',
                                    ylabel='Dilution Length (m)')


class Test_Han_2013_fig7(unittest.TestCase):
    """
    Tests for concentration over normalized axial distance with various orifice diameters and constant pressure, data from Han et al. 2013
    """
    def setUp(self):
        data = utils.read_csv(os.path.join(DATA_LOC, 'han-2013-fig7.csv'), header=2)

        with open(LIMITS_FILE) as f:
            self.limits = json.load(f)[__class__.__name__]

        # Set parameters
        temp = 293 #K
        pressure = 300 * bar
        amb_temp = 293 #K
        amb_pressure = 101325 #Pa

        air = Fluid(T=amb_temp, P=amb_pressure, species='air')

        self.diameters = np.array([0.5, 0.7, 1]) * milli #m

        self.output_data = dict()
        gas = Fluid(T=temp, P=pressure, species='h2')
        for idx, diameter in enumerate(self.diameters):
            orifice = Orifice(d=diameter)
            jet = Jet(gas, orifice, air, theta0=0, Ymin=1e-4)

            calc_distance = jet.S / diameter
            calc_concentration = 1 / jet.X_cl

            # Data is in x1/y1, x4/y4, x7/y7
            exp_distance = data[f'x{idx*3+1}']
            exp_concentration = data[f'y{idx*3+1}']

            self.output_data[diameter] = utils.interp_and_trim_data(calculated_x=calc_distance, calculated_y=calc_concentration,
                                                                    experiment_x=exp_distance, experiment_y=exp_concentration,
                                                                    upper_bound=15000)

    def test_centerline_concentration(self):
        for diameter in self.diameters:
            title = 'Han et al. 2013 - Figure 7, Inverse Mole Fraction vs. Axial Distance, d = ' + str(diameter / milli) + 'mm'

            data = self.output_data[diameter]
            error_limits = self.limits[title]
            error = utils.get_error(exp_vals=data['exp_y'],
                                    calc_vals=data['calc_y'],
                                    error_limits=error_limits,
                                    units='',
                                    msg=title,
                                    output_filename=OUTPUT_FILE,
                                    create_output=CREATE_OUTPUT,
                                    verbose=VERBOSE)

            with self.subTest():
                self.assertLessEqual(error['Max Absolute Error'], error_limits['Max Absolute Error'], f"{title}: Maximum Absolute Error out of range")
            with self.subTest():
                self.assertLessEqual(error['Avg Absolute Error'], error_limits['Avg Absolute Error'], f"{title}: Average Absolute Error out of range")
            with self.subTest():
                self.assertLessEqual(error['Max Percent Error'], error_limits['Max Percent Error'], f"{title}: Maximum Percent Error out of range")
            with self.subTest():
                self.assertLessEqual(error['Avg Percent Error'], error_limits['Avg Percent Error'], f"{title}: Average Percent Error out of range")
            with self.subTest():
                self.assertGreaterEqual(error['R2'], error_limits['R2'], f"{title}: R Squared Value out of range")

            if CREATE_PLOTS:
                utils.create_plots(output_dir=OUTPUT_LOC,
                                x=data['calc_x'],
                                exp_y=data['exp_y'],
                                calc_y=data['calc_y'],
                                max_error=error['Max Absolute Error'],
                                avg_error=error['Avg Absolute Error'],
                                smape=error['SMAPE'],
                                title=title,
                                xlabel='Normalized Axial Distance (z/D)',
                                ylabel='Inverse Mole Fraction')


class Test_Han_2013_fig8(unittest.TestCase):
    """
    Tests for concentration over normalized axial distance with constant orifice diameter and various pressures, data from Han et al. 2013
    """
    def setUp(self):
        data_a = utils.read_csv(os.path.join(DATA_LOC, 'han-2013-fig8a.csv'), header=2)
        data_b = utils.read_csv(os.path.join(DATA_LOC, 'han-2013-fig8b.csv'), header=2)
        data_c = utils.read_csv(os.path.join(DATA_LOC, 'han-2013-fig8c.csv'), header=2)
        data_d = utils.read_csv(os.path.join(DATA_LOC, 'han-2013-fig8d.csv'), header=2)

        csv_data = [data_a, data_b, data_c, data_d]

        with open(LIMITS_FILE) as f:
            self.limits = json.load(f)[__class__.__name__]

        # Set parameters
        temp = 293 #K
        pressure = 300 * bar
        amb_temp = 293 #K
        amb_pressure = 101325 #Pa

        air = Fluid(T=amb_temp, P=amb_pressure, species='air')

        diameter = 0.7 * milli
        self.pressures = np.arange(100, 401, 100) * bar

        self.output_data = dict()
        for idx, pressure in enumerate(self.pressures):
            gas = Fluid(T=temp, P=pressure, species='h2')
            orifice = Orifice(d=diameter)
            jet = Jet(gas, orifice, air, theta0=0, Ymin=1e-4)

            calc_distance = jet.S / diameter
            calc_concentration = 1 / jet.X_cl

            exp_distance = csv_data[idx]['x1']
            exp_concentration = csv_data[idx]['y1']

            self.output_data[pressure] = utils.interp_and_trim_data(calculated_x=calc_distance, calculated_y=calc_concentration,
                                                                    experiment_x=exp_distance, experiment_y=exp_concentration)

    def test_centerline_concentration(self):
        for pressure in self.pressures:
            title = 'Han et al. 2013 - Figure 8, Inverse Mole Fraction vs. Axial Distance, P0 = ' + str(pressure / bar) + ' bar'

            data = self.output_data[pressure]
            error_limits = self.limits[title]
            error = utils.get_error(exp_vals=data['exp_y'],
                                    calc_vals=data['calc_y'],
                                    error_limits=error_limits,
                                    units='',
                                    msg=title,
                                    output_filename=OUTPUT_FILE,
                                    create_output=CREATE_OUTPUT,
                                    verbose=VERBOSE)

            with self.subTest():
                self.assertLessEqual(error['Max Absolute Error'], error_limits['Max Absolute Error'], f"{title}: Maximum Absolute Error out of range")
            with self.subTest():
                self.assertLessEqual(error['Avg Absolute Error'], error_limits['Avg Absolute Error'], f"{title}: Average Absolute Error out of range")
            with self.subTest():
                self.assertLessEqual(error['Max Percent Error'], error_limits['Max Percent Error'], f"{title}: Maximum Percent Error out of range")
            with self.subTest():
                self.assertLessEqual(error['Avg Percent Error'], error_limits['Avg Percent Error'], f"{title}: Average Percent Error out of range")
            with self.subTest():
                self.assertGreaterEqual(error['R2'], error_limits['R2'], f"{title}: R Squared Value out of range")

            if CREATE_PLOTS:
                utils.create_plots(output_dir=OUTPUT_LOC,
                                x=data['calc_x'],
                                exp_y=data['exp_y'],
                                calc_y=data['calc_y'],
                                max_error=error['Max Absolute Error'],
                                avg_error=error['Avg Absolute Error'],
                                smape=error['SMAPE'],
                                title=title,
                                xlabel='Normalized Axial Distance (z/D)',
                                ylabel='Inverse Mole Fraction')

if __name__ == "__main__":
    if os.path.exists(OUTPUT_FILE):
        os.remove(OUTPUT_FILE)
    unittest.main()



"""
Copyright 2015-2023 National Technology & Engineering Solutions of Sandia, LLC (NTESS).
Under the terms of Contract DE-NA0003525 with NTESS, the U.S. Government retains certain rights in this software.

You should have received a copy of the GNU General Public License along with HyRAM+.
If not, see https://www.gnu.org/licenses/.
"""


def calc_pll(frequency, consequence):
    """
    Calculation of potential loss of life (PLL)

    Parameters
    ----------
    frequency : float
        Frequency of occurrence (events per year)
    
    consequence : float
        Fatal consequences of occurrence (fatalities)
    
    Returns
    -------
    pll : float
        Potential loss of life (fatalities per year)
    """
    pll = frequency * consequence
    return pll


def calc_far(pll, total_occupants):
    """
    Calculation of fatal accident rate (FAR)

    Parameters
    ----------
    pll : float
        Potential loss of life (fatalities per year)
    
    total_occupants : float
        Number of exposed people (people)
    
    Returns
    -------
    far : float
        Fatal accident rate (fatalities per 10^8 person*hours)
    """
    hours_per_year = 8760  # based on 24 hours per day, 365 days per year
    if pll == 0:
        far = 0
    else:
        far = (pll * 1e8) / (total_occupants * hours_per_year)
    return far


def calc_air(far, exposed_hours_per_year):
    """
    Calculation of average individual risk (AIR)

    Parameters
    ----------
    far : float
        Fatal accident rate (fatalities per 10^8 person*hours)
    
    exposed_hours_per_year : float
        Number of hours the individual spends in the facility,
        e.g., 2,000 hours for a full time worker
        (40 hours per week, 50 weeks per year)
    
    Returns
    -------
    air : float
        Average individual risk (fatalities per year)
    """
    air = far * 1e-8 * exposed_hours_per_year
    return air


def calc_all_plls(frequencies, consequences):
    """
    Calculation of multiple PLL metrics for multiple
    frequency and fatality values

    Parameters
    ----------
    frequencies : list
        List of frequency values (units of per year)
    
    consequences : list
        List of fatality values (units of fatalities)
    
    Returns
    -------
    plls : list
        List of PLL values (units of fatalities per year)
    """
    plls = []
    for frequency, consequence in zip(frequencies, consequences):
        pll = calc_pll(frequency, consequence)
        plls.append(pll)
    return plls


def calc_risk_contributions(risk_values):
    """
    Calculation of contribution to total risk from multiple risk values
    e.g., PLL

    Parameters
    ----------
    risk_values : list
        List of risk values
    
    Returns
    -------
    total_risk : float
        Sum total of risk from all inputs
    
    risk_contributions : list
        List of fractional contributions for each risk value
        in same order as input list
    """
    total_risk = sum(risk_values)
    risk_contributions = []
    for risk_value in risk_values:
        if total_risk > 0:
            risk_contribution = risk_value / total_risk
        else:
            risk_contribution = 0
        risk_contributions.append(risk_contribution)
    return total_risk, risk_contributions



"""
Copyright 2015-2023 National Technology & Engineering Solutions of Sandia, LLC (NTESS).
Under the terms of Contract DE-NA0003525 with NTESS, the U.S. Government retains certain rights in this software.

You should have received a copy of the GNU General Public License along with HyRAM+.
If not, see https://www.gnu.org/licenses/.
"""

import os
import unittest
import json

import numpy as np

from hyram.phys import Fluid, Orifice, Jet
from hyram.utilities import misc_utils

import utils

"""
NOTE: if running from IDE like pycharm, make sure cwd is hyram/ and not hyram/tests.
"""

# Flags to enable command line output, pyplots, and text file output
VERBOSE = False
CREATE_PLOTS = False
CREATE_OUTPUT = False

# Absolute paths to input/output data
DATA_LOC = os.path.join(os.path.dirname(__file__), 'data')
OUTPUT_LOC = os.path.join(misc_utils.get_temp_folder(), 'validation-cryo_concentrations')
OUTPUT_FILE = os.path.join(OUTPUT_LOC, 'Validation.txt')
LIMITS_FILE = os.path.join(DATA_LOC, 'Limits-cryo_concentrations.json')


class Test_Xiao_2010(unittest.TestCase):
    """
    Tests for table 5-3 from Molkov
    """
    def setUp(self):
        data = utils.read_csv(os.path.join(DATA_LOC, 'xiao-data.csv'), header=1)

        with open(LIMITS_FILE) as f:
            self.limits = json.load(f)[__class__.__name__]

        case1 = {'P': 1.7e6, 'd': 2.0e-3, 'T': 298}
        case2 = {'P': 6.85e6, 'd': 1.0e-3, 'T': 298}
        case3 = {'P': 0.825e6, 'd': 2.0e-3, 'T': 80}
        case4 = {'P': 3.2e6, 'd': 1.0e-3,'T': 80}

        self.output_data = []
        for idx, case in enumerate([case1, case2, case3, case4]):
            fluid = Fluid(T=case['T'], P=case['P'], species='h2')
            orifice = Orifice(case['d'])
            air = Fluid(T=295, P=101325, species='air')
            jet = Jet(fluid=fluid, orifice=orifice, ambient=air, Smax=6)

            calc_distance = jet.S / jet.developing_flow.d0
            calc_concentration = 1 / jet.X_cl

            case_num = str(idx + 1)
            exp_distance = data['case' + case_num + '_x']
            exp_concentration = data['case' + case_num + '_y']

            trimmed_data = utils.interp_and_trim_data(calculated_x=calc_distance, calculated_y=calc_concentration,
                                                      experiment_x=exp_distance, experiment_y=exp_concentration)

            self.output_data.append(trimmed_data)

    def test_concentration_1(self):
        case = self.output_data[0]
        title = "Xiao et al. 2010 - Figure 3, Inverse Mass Fraction over Normalized Distance, Case 1"

        error_limits = self.limits[title]
        error = utils.get_error(exp_vals=case['exp_y'],
                                calc_vals=case['calc_y'],
                                error_limits=error_limits,
                                units='',
                                msg=title,
                                output_filename=OUTPUT_FILE,
                                create_output=CREATE_OUTPUT,
                                verbose=VERBOSE)

        with self.subTest():
            self.assertLessEqual(error['Max Absolute Error'], error_limits['Max Absolute Error'], f"{title}: Maximum Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Absolute Error'], error_limits['Avg Absolute Error'], f"{title}: Average Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Max Percent Error'], error_limits['Max Percent Error'], f"{title}: Maximum Percent Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Percent Error'], error_limits['Avg Percent Error'], f"{title}: Average Percent Error out of range")
        with self.subTest():
            self.assertGreaterEqual(error['R2'], error_limits['R2'], f"{title}: R Squared Value out of range")

        if CREATE_PLOTS:
            utils.create_plots(output_dir=OUTPUT_LOC,
                            x=case['calc_x'],
                            exp_y=case['exp_y'],
                            calc_y=case['calc_y'],
                            max_error=error['Max Absolute Error'],
                            avg_error=error['Avg Absolute Error'],
                            smape=error['SMAPE'],
                            title=title,
                            xlabel=r'Normalized Distance ($S/D_0$)',
                            ylabel='Inverse Centerline Mole Fraction')

    def test_concentration_2(self):
        case = self.output_data[1]
        title = "Xiao et al. 2010 - Figure 3, Inverse Mass Fraction over Normalized Distance, Case 2"

        error_limits = self.limits[title]
        error = utils.get_error(exp_vals=case['exp_y'],
                                calc_vals=case['calc_y'],
                                error_limits=error_limits,
                                units='',
                                msg=title,
                                output_filename=OUTPUT_FILE,
                                create_output=CREATE_OUTPUT,
                                verbose=VERBOSE)

        with self.subTest():
            self.assertLessEqual(error['Max Absolute Error'], error_limits['Max Absolute Error'], f"{title}: Maximum Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Absolute Error'], error_limits['Avg Absolute Error'], f"{title}: Average Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Max Percent Error'], error_limits['Max Percent Error'], f"{title}: Maximum Percent Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Percent Error'], error_limits['Avg Percent Error'], f"{title}: Average Percent Error out of range")
        with self.subTest():
            self.assertGreaterEqual(error['R2'], error_limits['R2'], f"{title}: R Squared Value out of range")

        if CREATE_PLOTS:
            utils.create_plots(output_dir=OUTPUT_LOC,
                            x=case['calc_x'],
                            exp_y=case['exp_y'],
                            calc_y=case['calc_y'],
                            max_error=error['Max Absolute Error'],
                            avg_error=error['Avg Absolute Error'],
                            smape=error['SMAPE'],
                            title=title,
                            xlabel=r'Normalized Distance ($S/D_0$)',
                            ylabel='Inverse Centerline Mole Fraction')

    def test_concentration_3(self):
        case = self.output_data[2]
        title = "Xiao et al. 2010 - Figure 3, Inverse Mass Fraction over Normalized Distance, Case 3"

        error_limits = self.limits[title]
        error = utils.get_error(exp_vals=case['exp_y'],
                                calc_vals=case['calc_y'],
                                error_limits=error_limits,
                                units='',
                                msg=title,
                                output_filename=OUTPUT_FILE,
                                create_output=CREATE_OUTPUT,
                                verbose=VERBOSE)

        with self.subTest():
            self.assertLessEqual(error['Max Absolute Error'], error_limits['Max Absolute Error'], f"{title}: Maximum Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Absolute Error'], error_limits['Avg Absolute Error'], f"{title}: Average Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Max Percent Error'], error_limits['Max Percent Error'], f"{title}: Maximum Percent Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Percent Error'], error_limits['Avg Percent Error'], f"{title}: Average Percent Error out of range")
        with self.subTest():
            self.assertGreaterEqual(error['R2'], error_limits['R2'], f"{title}: R Squared Value out of range")

        if CREATE_PLOTS:
            utils.create_plots(output_dir=OUTPUT_LOC,
                            x=case['calc_x'],
                            exp_y=case['exp_y'],
                            calc_y=case['calc_y'],
                            max_error=error['Max Absolute Error'],
                            avg_error=error['Avg Absolute Error'],
                            smape=error['SMAPE'],
                            title=title,
                            xlabel=r'Normalized Distance ($S/D_0$)',
                            ylabel='Inverse Centerline Mole Fraction')

    def test_concentration_4(self):
        case = self.output_data[3]
        title = "Xiao et al. 2010 - Figure 3, Inverse Mass Fraction over Normalized Distance, Case 4"

        error_limits = self.limits[title]
        error = utils.get_error(exp_vals=case['exp_y'],
                                calc_vals=case['calc_y'],
                                error_limits=error_limits,
                                units='',
                                msg=title,
                                output_filename=OUTPUT_FILE,
                                create_output=CREATE_OUTPUT,
                                verbose=VERBOSE)

        with self.subTest():
            self.assertLessEqual(error['Max Absolute Error'], error_limits['Max Absolute Error'], f"{title}: Maximum Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Absolute Error'], error_limits['Avg Absolute Error'], f"{title}: Average Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Max Percent Error'], error_limits['Max Percent Error'], f"{title}: Maximum Percent Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Percent Error'], error_limits['Avg Percent Error'], f"{title}: Average Percent Error out of range")
        with self.subTest():
            self.assertGreaterEqual(error['R2'], error_limits['R2'], f"{title}: R Squared Value out of range")

        if CREATE_PLOTS:
            utils.create_plots(output_dir=OUTPUT_LOC,
                            x=case['calc_x'],
                            exp_y=case['exp_y'],
                            calc_y=case['calc_y'],
                            max_error=error['Max Absolute Error'],
                            avg_error=error['Avg Absolute Error'],
                            smape=error['SMAPE'],
                            title=title,
                            xlabel=r'Normalized Distance ($S/D_0$)',
                            ylabel='Inverse Centerline Mole Fraction')


class Test_Friedrich_Dist1(unittest.TestCase):
    """
    Tests for Inverse Centerline Mole Fraction vs. Normalized Distance
    'Normalized Distance' here is defined as z/d * sqrt(rho(ambient)/rho(initial))
    """
    def setUp(self):
        data = utils.read_csv(os.path.join(DATA_LOC, 'friedrich-fig5.csv'), header=1)

        with open(LIMITS_FILE) as f:
            self.limits = json.load(f)[__class__.__name__]

        experiments = ['3000', '3002', '3004', '3005', '5001', '5002']
        parameters = [{'P':19, 'T':37, 'd':1},
                       {'P':15, 'T':38, 'd':1},
                       {'P':29, 'T':36, 'd':1},
                       {'P':18, 'T':36, 'd':1},
                       {'P':29.85, 'T':43.59, 'd':0.5},
                       {'P':29.85, 'T':43.59, 'd':0.5}]

        self.output_data = []
        for idx, params in enumerate(parameters):
            fluid = Fluid(T=params['T'], P=params['P']*1e5, species='h2')
            orifice = Orifice(params['d']*1e-3)
            air = Fluid(T=295, P=101325, species='air')
            jet = Jet(fluid=fluid, orifice=orifice, ambient=air)

            calc_distance = jet.S / jet.developing_flow.d0 * (np.sqrt(jet.ambient.rho / jet.fluid.rho))
            calc_concentration = 1 / jet.X_cl

            case_num = experiments[idx]
            exp_distance = data['x_' + case_num]
            exp_concentration = np.array(data['y_' + case_num]) * 100

            trimmed_data = utils.interp_and_trim_data(calculated_x=calc_distance, calculated_y=calc_concentration,
                                                      experiment_x=exp_distance, experiment_y=exp_concentration)

            self.output_data.append(trimmed_data)

    def test_concentration_exp1(self):
        case = self.output_data[0]
        title = "Friedrich - Figure 5, Inverse Centerline Mole Fraction over Normalized Distance 1, Experiment #3000"

        error_limits = self.limits[title]
        error = utils.get_error(exp_vals=case['exp_y'],
                                calc_vals=case['calc_y'],
                                error_limits=error_limits,
                                units='',
                                msg=title,
                                output_filename=OUTPUT_FILE,
                                create_output=CREATE_OUTPUT,
                                verbose=VERBOSE)

        with self.subTest():
            self.assertLessEqual(error['Max Absolute Error'], error_limits['Max Absolute Error'], f"{title}: Maximum Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Absolute Error'], error_limits['Avg Absolute Error'], f"{title}: Average Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Max Percent Error'], error_limits['Max Percent Error'], f"{title}: Maximum Percent Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Percent Error'], error_limits['Avg Percent Error'], f"{title}: Average Percent Error out of range")
        with self.subTest():
            self.assertGreaterEqual(error['R2'], error_limits['R2'], f"{title}: R Squared Value out of range")

        if CREATE_PLOTS:
            utils.create_plots(output_dir=OUTPUT_LOC,
                                x=case['calc_x'],
                                exp_y=case['exp_y'],
                                calc_y=case['calc_y'],
                                max_error=error['Max Absolute Error'],
                                avg_error=error['Avg Absolute Error'],
                                smape=error['SMAPE'],
                                title=title,
                                xlabel=r'Normalized Distance $\left(\frac{z}{d}\sqrt{\frac{\rho_{\rm amb}}{\rho_{0}}}\right)$',
                                ylabel='Inverse Centerline Mole Fraction')

    def test_concentration_exp2(self):
        case = self.output_data[1]
        title = "Friedrich - Figure 5, Inverse Centerline Mole Fraction over Normalized Distance 1, Experiment #3002"

        error_limits = self.limits[title]
        error = utils.get_error(exp_vals=case['exp_y'],
                                calc_vals=case['calc_y'],
                                error_limits=error_limits,
                                units='',
                                msg=title,
                                output_filename=OUTPUT_FILE,
                                create_output=CREATE_OUTPUT,
                                verbose=VERBOSE)

        with self.subTest():
            self.assertLessEqual(error['Max Absolute Error'], error_limits['Max Absolute Error'], f"{title}: Maximum Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Absolute Error'], error_limits['Avg Absolute Error'], f"{title}: Average Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Max Percent Error'], error_limits['Max Percent Error'], f"{title}: Maximum Percent Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Percent Error'], error_limits['Avg Percent Error'], f"{title}: Average Percent Error out of range")
        with self.subTest():
            self.assertGreaterEqual(error['R2'], error_limits['R2'], f"{title}: R Squared Value out of range")

        if CREATE_PLOTS:
            utils.create_plots(output_dir=OUTPUT_LOC,
                                x=case['calc_x'],
                                exp_y=case['exp_y'],
                                calc_y=case['calc_y'],
                                max_error=error['Max Absolute Error'],
                                avg_error=error['Avg Absolute Error'],
                                smape=error['SMAPE'],
                                title=title,
                                xlabel=r'Normalized Distance $\left(\frac{z}{d}\sqrt{\frac{\rho_{\rm amb}}{\rho_{0}}}\right)$',
                                ylabel='Inverse Centerline Mole Fraction')

    def test_concentration_exp3(self):
        case = self.output_data[2]
        title = "Friedrich - Figure 5, Inverse Centerline Mole Fraction over Normalized Distance 1, Experiment #3004"

        error_limits = self.limits[title]
        error = utils.get_error(exp_vals=case['exp_y'],
                                calc_vals=case['calc_y'],
                                error_limits=error_limits,
                                units='',
                                msg=title,
                                output_filename=OUTPUT_FILE,
                                create_output=CREATE_OUTPUT,
                                verbose=VERBOSE)

        with self.subTest():
            self.assertLessEqual(error['Max Absolute Error'], error_limits['Max Absolute Error'], f"{title}: Maximum Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Absolute Error'], error_limits['Avg Absolute Error'], f"{title}: Average Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Max Percent Error'], error_limits['Max Percent Error'], f"{title}: Maximum Percent Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Percent Error'], error_limits['Avg Percent Error'], f"{title}: Average Percent Error out of range")
        with self.subTest():
            self.assertGreaterEqual(error['R2'], error_limits['R2'], f"{title}: R Squared Value out of range")

        if CREATE_PLOTS:
            utils.create_plots(output_dir=OUTPUT_LOC,
                                x=case['calc_x'],
                                exp_y=case['exp_y'],
                                calc_y=case['calc_y'],
                                max_error=error['Max Absolute Error'],
                                avg_error=error['Avg Absolute Error'],
                                smape=error['SMAPE'],
                                title=title,
                                xlabel=r'Normalized Distance $\left(\frac{z}{d}\sqrt{\frac{\rho_{\rm amb}}{\rho_{0}}}\right)$',
                                ylabel='Inverse Centerline Mole Fraction')

    def test_concentration_exp4(self):
        case = self.output_data[3]
        title = "Friedrich - Figure 5, Inverse Centerline Mole Fraction over Normalized Distance 1, Experiment #3005"

        error_limits = self.limits[title]
        error = utils.get_error(exp_vals=case['exp_y'],
                                calc_vals=case['calc_y'],
                                error_limits=error_limits,
                                units='',
                                msg=title,
                                output_filename=OUTPUT_FILE,
                                create_output=CREATE_OUTPUT,
                                verbose=VERBOSE)

        with self.subTest():
            self.assertLessEqual(error['Max Absolute Error'], error_limits['Max Absolute Error'], f"{title}: Maximum Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Absolute Error'], error_limits['Avg Absolute Error'], f"{title}: Average Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Max Percent Error'], error_limits['Max Percent Error'], f"{title}: Maximum Percent Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Percent Error'], error_limits['Avg Percent Error'], f"{title}: Average Percent Error out of range")
        with self.subTest():
            self.assertGreaterEqual(error['R2'], error_limits['R2'], f"{title}: R Squared Value out of range")

        if CREATE_PLOTS:
            utils.create_plots(output_dir=OUTPUT_LOC,
                                x=case['calc_x'],
                                exp_y=case['exp_y'],
                                calc_y=case['calc_y'],
                                max_error=error['Max Absolute Error'],
                                avg_error=error['Avg Absolute Error'],
                                smape=error['SMAPE'],
                                title=title,
                                xlabel=r'Normalized Distance $\left(\frac{z}{d}\sqrt{\frac{\rho_{\rm amb}}{\rho_{0}}}\right)$',
                                ylabel='Inverse Centerline Mole Fraction')

    def test_concentration_exp5(self):
        case = self.output_data[4]
        title = "Friedrich - Figure 5, Inverse Centerline Mole Fraction over Normalized Distance 1, Experiment #5001"

        error_limits = self.limits[title]
        error = utils.get_error(exp_vals=case['exp_y'],
                                calc_vals=case['calc_y'],
                                error_limits=error_limits,
                                units='',
                                msg=title,
                                output_filename=OUTPUT_FILE,
                                create_output=CREATE_OUTPUT,
                                verbose=VERBOSE)

        with self.subTest():
            self.assertLessEqual(error['Max Absolute Error'], error_limits['Max Absolute Error'], f"{title}: Maximum Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Absolute Error'], error_limits['Avg Absolute Error'], f"{title}: Average Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Max Percent Error'], error_limits['Max Percent Error'], f"{title}: Maximum Percent Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Percent Error'], error_limits['Avg Percent Error'], f"{title}: Average Percent Error out of range")
        with self.subTest():
            self.assertGreaterEqual(error['R2'], error_limits['R2'], f"{title}: R Squared Value out of range")

        if CREATE_PLOTS:
            utils.create_plots(output_dir=OUTPUT_LOC,
                                x=case['calc_x'],
                                exp_y=case['exp_y'],
                                calc_y=case['calc_y'],
                                max_error=error['Max Absolute Error'],
                                avg_error=error['Avg Absolute Error'],
                                smape=error['SMAPE'],
                                title=title,
                                xlabel=r'Normalized Distance $\left(\frac{z}{d}\sqrt{\frac{\rho_{\rm amb}}{\rho_{0}}}\right)$',
                                ylabel='Inverse Centerline Mole Fraction')

    def test_concentration_exp6(self):
        case = self.output_data[5]
        title = "Friedrich - Figure 5, Inverse Centerline Mole Fraction over Normalized Distance 1, Experiment #5002"

        error_limits = self.limits[title]
        error = utils.get_error(exp_vals=case['exp_y'],
                                calc_vals=case['calc_y'],
                                error_limits=error_limits,
                                units='',
                                msg=title,
                                output_filename=OUTPUT_FILE,
                                create_output=CREATE_OUTPUT,
                                verbose=VERBOSE)

        with self.subTest():
            self.assertLessEqual(error['Max Absolute Error'], error_limits['Max Absolute Error'], f"{title}: Maximum Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Absolute Error'], error_limits['Avg Absolute Error'], f"{title}: Average Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Max Percent Error'], error_limits['Max Percent Error'], f"{title}: Maximum Percent Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Percent Error'], error_limits['Avg Percent Error'], f"{title}: Average Percent Error out of range")
        with self.subTest():
            self.assertGreaterEqual(error['R2'], error_limits['R2'], f"{title}: R Squared Value out of range")

        if CREATE_PLOTS:
            utils.create_plots(output_dir=OUTPUT_LOC,
                                x=case['calc_x'],
                                exp_y=case['exp_y'],
                                calc_y=case['calc_y'],
                                max_error=error['Max Absolute Error'],
                                avg_error=error['Avg Absolute Error'],
                                smape=error['SMAPE'],
                                title=title,
                                xlabel=r'Normalized Distance $\left(\frac{z}{d}\sqrt{\frac{\rho_{\rm amb}}{\rho_{0}}}\right)$',
                                ylabel='Inverse Centerline Mole Fraction')

class Test_Friedrich_Dist2(unittest.TestCase):
    """
    Tests for Inverse Centerline Mole Fraction vs. Normalized Distance
    'Normalized Distance' here is defined as z/d * sqrt(rho(ambient)/rho(nozzle))
    """
    def setUp(self):
        data = utils.read_csv(os.path.join(DATA_LOC, 'friedrich-fig5.csv'), header=1)

        with open(LIMITS_FILE) as f:
            self.limits = json.load(f)[__class__.__name__]

        experiments = ['3000', '3002', '3004', '3005', '5001', '5002']
        parameters = [{'P':19, 'T':37, 'd':1},
                       {'P':15, 'T':38, 'd':1},
                       {'P':29, 'T':36, 'd':1},
                       {'P':18, 'T':36, 'd':1},
                       {'P':29.85, 'T':43.59, 'd':0.5},
                       {'P':29.85, 'T':43.59, 'd':0.5}]

        self.output_data = []
        for idx, params in enumerate(parameters):
            fluid = Fluid(T=params['T'], P=params['P']*1e5, species='H2')
            orifice = Orifice(params['d']*1e-3)
            air = Fluid(T=295, P=101325, species='air')
            jet = Jet(fluid=fluid, orifice=orifice, ambient=air)

            calc_distance = jet.S / jet.developing_flow.d0 * (np.sqrt(jet.ambient.rho / jet.developing_flow.fluid_orifice.rho))
            calc_concentration = 1 / jet.X_cl

            case_num = experiments[idx]
            exp_distance = data['x_' + case_num]
            exp_concentration = np.array(data['y_' + case_num]) * 100

            trimmed_data = utils.interp_and_trim_data(calculated_x=calc_distance, calculated_y=calc_concentration,
                                                      experiment_x=exp_distance, experiment_y=exp_concentration)

            self.output_data.append(trimmed_data)

    def test_concentration_exp1(self):
        case = self.output_data[0]
        title = "Friedrich - Figure 5, Inverse Centerline Mole Fraction over Normalized Distance 2, Experiment #3000"

        error_limits = self.limits[title]
        error = utils.get_error(exp_vals=case['exp_y'],
                                calc_vals=case['calc_y'],
                                error_limits=error_limits,
                                units='',
                                msg=title,
                                output_filename=OUTPUT_FILE,
                                create_output=CREATE_OUTPUT,
                                verbose=VERBOSE)

        with self.subTest():
            self.assertLessEqual(error['Max Absolute Error'], error_limits['Max Absolute Error'], f"{title}: Maximum Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Absolute Error'], error_limits['Avg Absolute Error'], f"{title}: Average Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Max Percent Error'], error_limits['Max Percent Error'], f"{title}: Maximum Percent Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Percent Error'], error_limits['Avg Percent Error'], f"{title}: Average Percent Error out of range")
        with self.subTest():
            self.assertGreaterEqual(error['R2'], error_limits['R2'], f"{title}: R Squared Value out of range")

        if CREATE_PLOTS:
            utils.create_plots(output_dir=OUTPUT_LOC,
                                x=case['calc_x'],
                                exp_y=case['exp_y'],
                                calc_y=case['calc_y'],
                                max_error=error['Max Absolute Error'],
                                avg_error=error['Avg Absolute Error'],
                                smape=error['SMAPE'],
                                title=title,
                                xlabel=r'Normalized Distance $\left(\frac{z}{d}\sqrt{\frac{\rho_{\rm amb}}{\rho_{0}}}\right)$',
                                ylabel='Inverse Centerline Mole Fraction')

    def test_concentration_exp2(self):
        case = self.output_data[1]
        title = "Friedrich - Figure 5, Inverse Centerline Mole Fraction over Normalized Distance 2, Experiment #3002"

        error_limits = self.limits[title]
        error = utils.get_error(exp_vals=case['exp_y'],
                                calc_vals=case['calc_y'],
                                error_limits=error_limits,
                                units='',
                                msg=title,
                                output_filename=OUTPUT_FILE,
                                create_output=CREATE_OUTPUT,
                                verbose=VERBOSE)

        with self.subTest():
            self.assertLessEqual(error['Max Absolute Error'], error_limits['Max Absolute Error'], f"{title}: Maximum Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Absolute Error'], error_limits['Avg Absolute Error'], f"{title}: Average Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Max Percent Error'], error_limits['Max Percent Error'], f"{title}: Maximum Percent Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Percent Error'], error_limits['Avg Percent Error'], f"{title}: Average Percent Error out of range")
        with self.subTest():
            self.assertGreaterEqual(error['R2'], error_limits['R2'], f"{title}: R Squared Value out of range")

        if CREATE_PLOTS:
            utils.create_plots(output_dir=OUTPUT_LOC,
                                x=case['calc_x'],
                                exp_y=case['exp_y'],
                                calc_y=case['calc_y'],
                                max_error=error['Max Absolute Error'],
                                avg_error=error['Avg Absolute Error'],
                                smape=error['SMAPE'],
                                title=title,
                                xlabel=r'Normalized Distance $\left(\frac{z}{d}\sqrt{\frac{\rho_{\rm amb}}{\rho_{0}}}\right)$',
                                ylabel='Inverse Centerline Mole Fraction')

    def test_concentration_exp3(self):
        case = self.output_data[2]
        title = "Friedrich - Figure 5, Inverse Centerline Mole Fraction over Normalized Distance 2, Experiment #3004"

        error_limits = self.limits[title]
        error = utils.get_error(exp_vals=case['exp_y'],
                                calc_vals=case['calc_y'],
                                error_limits=error_limits,
                                units='',
                                msg=title,
                                output_filename=OUTPUT_FILE,
                                create_output=CREATE_OUTPUT,
                                verbose=VERBOSE)

        with self.subTest():
            self.assertLessEqual(error['Max Absolute Error'], error_limits['Max Absolute Error'], f"{title}: Maximum Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Absolute Error'], error_limits['Avg Absolute Error'], f"{title}: Average Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Max Percent Error'], error_limits['Max Percent Error'], f"{title}: Maximum Percent Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Percent Error'], error_limits['Avg Percent Error'], f"{title}: Average Percent Error out of range")
        with self.subTest():
            self.assertGreaterEqual(error['R2'], error_limits['R2'], f"{title}: R Squared Value out of range")

        if CREATE_PLOTS:
            utils.create_plots(output_dir=OUTPUT_LOC,
                                x=case['calc_x'],
                                exp_y=case['exp_y'],
                                calc_y=case['calc_y'],
                                max_error=error['Max Absolute Error'],
                                avg_error=error['Avg Absolute Error'],
                                smape=error['SMAPE'],
                                title=title,
                                xlabel=r'Normalized Distance $\left(\frac{z}{d}\sqrt{\frac{\rho_{\rm amb}}{\rho_{0}}}\right)$',
                                ylabel='Inverse Centerline Mole Fraction')

    def test_concentration_exp4(self):
        case = self.output_data[3]
        title = "Friedrich - Figure 5, Inverse Centerline Mole Fraction over Normalized Distance 2, Experiment #3005"

        error_limits = self.limits[title]
        error = utils.get_error(exp_vals=case['exp_y'],
                                calc_vals=case['calc_y'],
                                error_limits=error_limits,
                                units='',
                                msg=title,
                                output_filename=OUTPUT_FILE,
                                create_output=CREATE_OUTPUT,
                                verbose=VERBOSE)

        with self.subTest():
            self.assertLessEqual(error['Max Absolute Error'], error_limits['Max Absolute Error'], f"{title}: Maximum Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Absolute Error'], error_limits['Avg Absolute Error'], f"{title}: Average Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Max Percent Error'], error_limits['Max Percent Error'], f"{title}: Maximum Percent Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Percent Error'], error_limits['Avg Percent Error'], f"{title}: Average Percent Error out of range")
        with self.subTest():
            self.assertGreaterEqual(error['R2'], error_limits['R2'], f"{title}: R Squared Value out of range")

        if CREATE_PLOTS:
            utils.create_plots(output_dir=OUTPUT_LOC,
                                x=case['calc_x'],
                                exp_y=case['exp_y'],
                                calc_y=case['calc_y'],
                                max_error=error['Max Absolute Error'],
                                avg_error=error['Avg Absolute Error'],
                                smape=error['SMAPE'],
                                title=title,
                                xlabel=r'Normalized Distance $\left(\frac{z}{d}\sqrt{\frac{\rho_{\rm amb}}{\rho_{0}}}\right)$',
                                ylabel='Inverse Centerline Mole Fraction')

    def test_concentration_exp5(self):
        case = self.output_data[4]
        title = "Friedrich - Figure 5, Inverse Centerline Mole Fraction over Normalized Distance 2, Experiment #5001"

        error_limits = self.limits[title]
        error = utils.get_error(exp_vals=case['exp_y'],
                                calc_vals=case['calc_y'],
                                error_limits=error_limits,
                                units='',
                                msg=title,
                                output_filename=OUTPUT_FILE,
                                create_output=CREATE_OUTPUT,
                                verbose=VERBOSE)

        with self.subTest():
            self.assertLessEqual(error['Max Absolute Error'], error_limits['Max Absolute Error'], f"{title}: Maximum Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Absolute Error'], error_limits['Avg Absolute Error'], f"{title}: Average Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Max Percent Error'], error_limits['Max Percent Error'], f"{title}: Maximum Percent Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Percent Error'], error_limits['Avg Percent Error'], f"{title}: Average Percent Error out of range")
        with self.subTest():
            self.assertGreaterEqual(error['R2'], error_limits['R2'], f"{title}: R Squared Value out of range")

        if CREATE_PLOTS:
            utils.create_plots(output_dir=OUTPUT_LOC,
                                x=case['calc_x'],
                                exp_y=case['exp_y'],
                                calc_y=case['calc_y'],
                                max_error=error['Max Absolute Error'],
                                avg_error=error['Avg Absolute Error'],
                                smape=error['SMAPE'],
                                title=title,
                                xlabel=r'Normalized Distance $\left(\frac{z}{d}\sqrt{\frac{\rho_{\rm amb}}{\rho_{0}}}\right)$',
                                ylabel='Inverse Centerline Mole Fraction')

    def test_concentration_exp6(self):
        case = self.output_data[5]
        title = "Friedrich - Figure 5, Inverse Centerline Mole Fraction over Normalized Distance 2, Experiment #5002"

        error_limits = self.limits[title]
        error = utils.get_error(exp_vals=case['exp_y'],
                                calc_vals=case['calc_y'],
                                error_limits=error_limits,
                                units='',
                                msg=title,
                                output_filename=OUTPUT_FILE,
                                create_output=CREATE_OUTPUT,
                                verbose=VERBOSE)

        with self.subTest():
            self.assertLessEqual(error['Max Absolute Error'], error_limits['Max Absolute Error'], f"{title}: Maximum Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Absolute Error'], error_limits['Avg Absolute Error'], f"{title}: Average Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Max Percent Error'], error_limits['Max Percent Error'], f"{title}: Maximum Percent Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Percent Error'], error_limits['Avg Percent Error'], f"{title}: Average Percent Error out of range")
        with self.subTest():
            self.assertGreaterEqual(error['R2'], error_limits['R2'], f"{title}: R Squared Value out of range")

        if CREATE_PLOTS:
            utils.create_plots(output_dir=OUTPUT_LOC,
                                x=case['calc_x'],
                                exp_y=case['exp_y'],
                                calc_y=case['calc_y'],
                                max_error=error['Max Absolute Error'],
                                avg_error=error['Avg Absolute Error'],
                                smape=error['SMAPE'],
                                title=title,
                                xlabel=r'Normalized Distance $\left(\frac{z}{d}\sqrt{\frac{\rho_{\rm amb}}{\rho_{0}}}\right)$',
                                ylabel='Inverse Centerline Mole Fraction')

if __name__ == "__main__":
    if os.path.exists(OUTPUT_FILE):
        os.remove(OUTPUT_FILE)
    unittest.main()



"""
Copyright 2015-2023 National Technology & Engineering Solutions of Sandia, LLC (NTESS).
Under the terms of Contract DE-NA0003525 with NTESS, the U.S. Government retains certain rights in this software.

You should have received a copy of the GNU General Public License along with HyRAM+.
If not, see https://www.gnu.org/licenses/.

"""
import unittest

from hyram.phys._therm import CoolPropWrapper, Combustion
from hyram.phys import Fluid


VERBOSE = False


class TestPropsSI(unittest.TestCase):
    def setUp(self):
        self.pressure = 1286500
        self.entropy = 10761.12726
        self.therm = CoolPropWrapper('H2')

    def test_wrapper_single(self):
        t = self.therm.PropsSI('T', P=self.pressure, S=self.entropy)
        self.assertTrue

    def test_wrapper_multi(self):
        h, rho = self.therm.PropsSI(['H', 'D'], P=self.pressure, S=self.entropy)
        self.assertTrue

    def test_wrapper_blend(self):
        therm = CoolPropWrapper("METHANE[0.5]&HYDROGEN[0.5]")
        t = therm.PropsSI('T', P=self.pressure, S=self.entropy)
        self.assertTrue

class TestCombustion(unittest.TestCase):
    def setUp(self):
        H2 = Fluid('H2', P = 101325, T = 298)
        CH4 = Fluid('CH4', P = 101325, T = 298)
        self.comb_H2 = Combustion(H2)
        self.comb_CH4 = Combustion(CH4)

    def test_H2_absorption_coeff(self):
        self.assertAlmostEqual(self.comb_H2.absorption_coeff, 0.23, None, "H2 absorption coefficient not 0.23", 0.02)

    def test_CH4_absorption_coeff(self):
        self.assertAlmostEqual(self.comb_CH4.absorption_coeff, 0.5, None, "CH4 absorption coefficient not 0.5", 0.02)

if __name__ == "__main__":
    unittest.main()


"""
Copyright 2015-2023 National Technology & Engineering Solutions of Sandia, LLC (NTESS).
Under the terms of Contract DE-NA0003525 with NTESS, the U.S. Government retains certain rights in this software.

You should have received a copy of the GNU General Public License along with HyRAM+.
If not, see https://www.gnu.org/licenses/.
"""

import unittest

from hyram.qra import risk


class TestRiskMetricCalcs(unittest.TestCase):
    """
    Test basic calculation of risk metrics
    """
    def test_calc_pll(self):
        frequency = 1.12e-5  # per year
        consequence = 0.1  # fatalties
        pll = risk.calc_pll(frequency, consequence)
        # Hand-calculation of above numbers
        self.assertAlmostEqual(pll, 1.12e-6)

    def test_calc_far(self):
        pll = 2.5e-5  # fatalties/year
        total_occupants = 9
        far = risk.calc_far(pll, total_occupants)
        # Hand-calculation of above numbers
        self.assertAlmostEqual(far, 0.0317098)

    def test_zero_far_for_zero_pll(self):
        pll = 0  # fatalties/year
        total_occupants = 0
        far = risk.calc_far(pll, total_occupants)
        # Hand-calculation of above numbers
        self.assertAlmostEqual(far, 0)

    def test_calc_air(self):
        far = 0.03  # fatalties per 10^8 person*hours
        exposed_hours_per_year = 2000
        air = risk.calc_air(far, exposed_hours_per_year)
        # Hand-calculation of above numbers
        self.assertAlmostEqual(air, 6e-7)


class TestScenarioRiskCalcs(unittest.TestCase):
    """
    Test calculation of risk (PLL) for all scenarios
    """
    def test_calc_all_plls(self):
        frequencies = [1e-5, 2e-5]
        fatalities = [0.1, 0.2]
        plls = risk.calc_all_plls(frequencies, fatalities)
        # Hand-calculation of above numbers
        hand_calc_plls = [1e-6, 4e-6]
        for calc_pll, test_pll in zip(plls, hand_calc_plls):
            self.assertAlmostEqual(calc_pll, test_pll)

    def test_calc_risk_contributions(self):
        plls = [1e-6, 4e-6]
        total_pll, pll_contributions = risk.calc_risk_contributions(plls)
        # Hand-calculation of above numbers
        self.assertAlmostEqual(total_pll, 5e-6)
        hand_calc_contributions = [0.2, 0.8]
        for calc_value, test_value in zip(pll_contributions, hand_calc_contributions):
            self.assertAlmostEqual(calc_value, test_value)

    def test_calc_risk_contributions_handles_zeros(self):
        plls = [0, 0]
        total_pll, pll_contributions = risk.calc_risk_contributions(plls)
        # Hand-calculation of above numbers
        self.assertAlmostEqual(total_pll, 0)
        hand_calc_contributions = [0, 0]
        for calc_value, test_value in zip(pll_contributions, hand_calc_contributions):
            self.assertAlmostEqual(calc_value, test_value)


if __name__ == "__main__":
    unittest.main()



"""
Copyright 2015-2023 National Technology & Engineering Solutions of Sandia, LLC (NTESS).
Under the terms of Contract DE-NA0003525 with NTESS, the U.S. Government retains certain rights in this software.

You should have received a copy of the GNU General Public License along with HyRAM+.
If not, see https://www.gnu.org/licenses/.
"""
import numpy as np
from .component_failure import ComponentFailureSet
from .component_set import ComponentSet


def init_leak_results(leak_sizes, f_release_overrides, component_sets, mass_flow_leak_size=None, mass_flow=None):
    """
    Creates LeakSizeResult objects for each leak size and obtains leak frequency data for each.
    Release frequencies are sum of frequencies for all components at that leak size.

    Parameters
    ----------
    leak_sizes : list(floats) or None
        leak sizes (% area of pipe)

    f_release_overrides : list(floats, None) or None
        Optional release frequency values which override those calculated by HyRAM

    component_sets : list(component_set)
        list of component set objects for each leak size.

    mass_flow_leak_size : {1, 10, 100, 1000, 10000}, optional
        Leak size (as hundredth of percent) of specified mass flow rate, assuming unchoked flow.

    mass_flow : float, optional
        [kg/s] Mass flow rate for specified leak size, if unchoked

    Returns
    -------
    leak_results: list(LeakSizeResult)
        list of leak size result objects ordered by size

    """
    if leak_sizes is None:
        leak_sizes = [0.01, 0.10, 1, 10, 100]

    # always included 100% size
    leak_sizes.sort()
    if leak_sizes[-1] != 100:
        leak_sizes.append(100)

    if f_release_overrides is None:
        f_release_overrides = [None] * len(leak_sizes)

    if mass_flow is not None:
        perc_leak_sizes = np.array([1, 10, 100, 1000, 10000])
        if mass_flow_leak_size not in perc_leak_sizes:
            raise ValueError(f'{mass_flow_leak_size} Leak size for unchoked mass flow not recognized')
        base_mass_flow = mass_flow / float(mass_flow_leak_size)
        mass_flows = base_mass_flow * perc_leak_sizes
    else:
        mass_flows = [None] * len(leak_sizes)

    leak_results = []
    for i, size in enumerate(leak_sizes):
        leak_result = LeakSizeResult(size, f_release_overrides[i], component_sets, mass_flow_override=mass_flows[i])
        leak_results.append(leak_result)

    return leak_results


class LeakSizeResult:
    """
    Represents leak release, including all components, at specified size.

    Parameters
    ----------
    leak_size : float
        Leak size as fraction of total area.

    f_release_override : float, optional
        Override indicating total frequency of release.

    component_sets : [ComponentSet]
        Component leak probability data for this leak size.

    mass_flow_override : float, optional
        Override indicating unchoked mass flow [kg/s].

    """
    leak_size: float
    mass_flow_rate: float or None = None
    mass_flow_override: float or None = None  # override for unchoked flow
    leak_diam: float or None = None
    f_release: float = 0  # Total release frequency for this leak size
    use_override: bool = False  # True if release frequency is set by user

    list_event_names: list
    list_event_keys: list
    list_p_events: list
    list_avg_events: list
    list_pll_contrib: list
    event_dicts: list  # list of dicts e.g. {id, label, prob, events, pll}
    f_component_leaks = {}

    # 100% Leak Size properties (None for all other sizes)
    f_failure: float or None = None  # Total vehicle & shutdown failure frequency
    use_failure_override: bool = False  # True if failure frequency is set by user
    failure_set: ComponentFailureSet or None = None

    def __init__(self, leak_size: float, f_release_override: float or None, component_sets: [ComponentSet], mass_flow_override=None):
        self.list_event_names = []
        self.list_event_keys = []
        self.list_p_events = []
        self.list_avg_events = []
        self.list_pll_contrib = []
        self.event_dicts = []
        self.f_component_leaks = {}

        self.leak_size = leak_size
        self.mass_flow_override = mass_flow_override

        # Calculate leak frequency from component releases or from override, if given.
        if f_release_override is not None:
            self.use_override = True
            self.f_release = f_release_override
            for comp_set in component_sets:
                self.f_component_leaks[comp_set.category] = 0

        else:
            for comp_set in component_sets:
                component_leak_frequency = comp_set.get_leak_frequency(self.leak_size)
                # print(f'size {self.leak_size}, comp {comp_set.category}, freq {component_leak_frequency}')
                self.f_component_leaks[comp_set.category] = component_leak_frequency
                self.f_release += component_leak_frequency

    def _set_rounded_attr(self, attr_name, val):
        """ Helper func to assign probability parameters """
        if val is not None:
            setattr(self, attr_name, np.around(val, 20))

    def set_failure_set(self, failure_set):
        """
        Updates failure frequency, release frequency, and related flags from failure set.
        NOTE: only used for 100% leak.

        Parameters
        ----------
        failure_set : ComponentFailureSet

        """
        self.failure_set = failure_set
        self.use_failure_override = failure_set.use_override

        self.f_failure = failure_set.f_failure  # equals override, if provided
        if not self.use_override:
            self.f_release += failure_set.f_failure

    def to_dict(self):
        """ Creates shallow copy of parameters without failure set class, for easy JSON serialization. """
        result = vars(self).copy()
        result['list_p_events'] = list(self.list_p_events)
        result['f_component_leaks'] = self.f_component_leaks

        # Pass through probabilities and frequencies of component failures, if an override isn't used.
        if self.failure_set is not None:
            if not self.use_failure_override:
                result |= {
                    'p_overp_rupture': self.failure_set.p_overp_rupture,
                    'f_overp_rupture': self.failure_set.f_overp_rupture,
                    'p_driveoff': self.failure_set.p_driveoff,
                    'f_driveoff': self.failure_set.f_driveoff,
                    'p_nozzle_release': self.failure_set.p_nozzle_release,
                    'f_nozzle_release': self.failure_set.f_nozzle_release,
                    'p_sol_valves_ftc': self.failure_set.p_sol_valves_ftc,
                    'f_sol_valves_ftc': self.failure_set.f_sol_valves_ftc,
                    'p_mvalve_ftc': self.failure_set.p_mvalve_ftc,
                    'f_mvalve_ftc': self.failure_set.f_mvalve_ftc
                }
            result.pop('failure_set', None)

        return result

    def sum_probabilities(self):
        return sum(self.list_p_events)

    def sum_events(self):
        return sum(self.list_avg_events)

    def sum_plls(self):
        return sum(self.list_pll_contrib)

    def get_result_dicts(self):
        """ Get dict of event data """
        result_dicts = []
        for i, name in enumerate(self.list_event_names):
            rdict = {'label': name,
                     'key': self.list_event_keys[i],
                     'prob': self.list_p_events[i],
                     'events': self.list_avg_events[i],
                     'pll': self.list_pll_contrib[i]}
            result_dicts.append(rdict)

        total_dict = {'label': 'TOTAL', 'key': 'tot',
                      'prob': self.sum_probabilities(),
                      'events': self.sum_events(), 'pll': self.sum_plls()}
        result_dicts.append(total_dict)
        return result_dicts

    def __repr__(self):
        return f"{self.leak_size:06.2f}% Leak Release"

    def __str__(self):
        # TODO (Cianan): does anyone use the table string generators below?
        override_str = "(override)" if self.use_override else ""
        failure_str = f", vehicle failure probabilities: {self.f_failure:.6}" if self.leak_size == 100 else ""
        msg = f"{self.leak_size:06.2f}% Leak release - total leak frequency {override_str} {self.f_release:.3g}{failure_str}"
        return msg

    # def _get_failure_prob_table(self) -> str:
    #     """ Generate vehicle failure probability string in tabular format """
    #     p_vehicle_fail_list = [
    #         {'label': 'Overpressure rupture', 'prob': self.failure_set.p_overp_rupture},
    #         {'label': 'Driveoff', 'prob': self.failure_set.p_driveoff},
    #         {'label': 'Nozzle release', 'prob': self.failure_set.p_nozzle_release},
    #         {'label': 'Solenoid valve FTC', 'prob': self.failure_set.p_sol_valves_ftc},
    #         {'label': 'Manual valve FTC', 'prob': self.failure_set.p_mvalve_ftc},
    #     ]
    #
    #     tmpl_hdr = "{label:<20} | {prob:>15}"
    #     str = '{}\n'.format(tmpl_hdr.format(label="Failure", prob="Probability"))
    #
    #     template = "{label:<20} | {prob:>15.10f}\n"
    #     for item in p_vehicle_fail_list:
    #         str += template.format(**item)
    #     return str

    # def _get_result_table(self) -> str:
    #     """ Generate leak result data string in tabular format """
    #     tmpl_hdr = "{label:<12} | {events:>15} | {prob:>15} | {pll:>15}"
    #     str = '{}\n'.format(tmpl_hdr.format(label="Type", events="Avg Events", prob="Branch line P", pll="PLL Contrib"))
    #
    #     template = "{label:<12} | {events:>15.10f} | {prob:>15.10f} | {pll:>15.10f}\n"
    #     for entry in self.get_result_dicts():
    #         str += template.format(**entry)
    #     return str

    # def _get_comp_freq_table(self) -> str:
    #     """ Generate component leak freq data string in tabular format """
    #     tmpl_hdr = "{label:<12} | {freq:>15}"
    #     str = '{}\n'.format(tmpl_hdr.format(label="Component", freq="Leak Freq"))
    #
    #     template = "{label:<12} | {freq:>15.10f}\n"
    #     for key, val in self.f_component_leaks.items():
    #         str += template.format(label=key, freq=val)
    #     return str



"""
Copyright 2015-2023 National Technology & Engineering Solutions of Sandia, LLC (NTESS).
Under the terms of Contract DE-NA0003525 with NTESS, the U.S. Government retains certain rights in this software.

You should have received a copy of the GNU General Public License along with HyRAM+.
If not, see https://www.gnu.org/licenses/.
"""


def get_ignition_probability(mass_flow_rate, mass_flow_thresholds, immed_ign_probs, delayed_ign_probs):
    """
    Get the ignition probabilities based on mass flow rate
    from a list immediate and delayed ignition probabilities
    that are separated into mass flow rate thresholds

    For the lists of ignition probabilities,
    the first value will be used if the input mass_flow_rate
    is less than the first (smallest) value of the mass_flow_thresholds,
    the last value will be used if the input mass_flow_rate
    is greater than the last (largest) value of the mass_flow_thresholds,
    and intermediate values will be used if the input mass_flow_rate
    falls within the values of the mass_flow_thresholds

    Parameters
    ----------
    mass_flow_rate : float
        Mass flow rate [kg/s] of interest

    mass_flow_thresholds : list of floats
        Sorted list of mass flow rates [kg/s] that separate the list of ignition probabilities

    immed_ign_probs : list of floats
        List of immediate ignition probabilities based on the mass flow rate thresholds 

    delayed_ign_probs : list of floats
        List of delayed ignition probabilities based on the mass flow rate thresholds

    Returns
    -------
    immediate_ignition_probability : float
        Probability of immediate ignition
    
    delayed_ignition_probability : float
        Probability of delayed ignition
    """
    if mass_flow_thresholds != sorted(mass_flow_thresholds):
        error_str = ('List of ignition thresholds ({})'.format(mass_flow_thresholds)
                     + ' must be ascending values for ignition probabilities.')
        raise ValueError(error_str)
    
    for prob_list in [immed_ign_probs, delayed_ign_probs]:
        if len(mass_flow_thresholds) != (len(prob_list) - 1):
            error_str = (
                'Number of ignition thresholds ({})'.format(len(mass_flow_thresholds))
                + ' must be exactly one value less than the'
                + ' number of ignition probabilities ({}).'.format(len(prob_list))
            )
            raise ValueError(error_str)

    if mass_flow_rate >= mass_flow_thresholds[-1]:
        threshold_index = len(mass_flow_thresholds)
    else:
        threshold_index = next(idx
                               for idx, max_threshold
                               in enumerate(mass_flow_thresholds)
                               if max_threshold > mass_flow_rate)
    immediate_ignition_probability = immed_ign_probs[threshold_index]
    delayed_ignition_probability = delayed_ign_probs[threshold_index]
    return immediate_ignition_probability, delayed_ignition_probability



"""
Copyright 2015-2023 National Technology & Engineering Solutions of Sandia, LLC (NTESS).
Under the terms of Contract DE-NA0003525 with NTESS, the U.S. Government retains certain rights in this software.

You should have received a copy of the GNU General Public License along with HyRAM+.
If not, see https://www.gnu.org/licenses/.
"""

import unittest

from hyram.qra import probits


class TestFatalityProbabilityCalc(unittest.TestCase):
    """
    Tests of calculation of fatality probability calculation
    """
    def test_zero_probability(self):
        value_much_lower_than_5 = -100
        self.assertAlmostEqual(probits.calculate_fatality_probability(value_much_lower_than_5), 0)

    def test_100pct_probability(self):
        value_much_grater_than_5 = 100
        self.assertAlmostEqual(probits.calculate_fatality_probability(value_much_grater_than_5), 1)

    def test_50pct_probability(self):
        self.assertAlmostEqual(probits.calculate_fatality_probability(5), 0.5)


class TestThermalProbits(unittest.TestCase):
    """
    Tests of thermal probit calculations
    """
    def test_calculate_thermal_dose(self):
        # Hand-calculation comparison
        heat_flux = 5  # W/m^2
        exposure_time = 30  # seconds
        self.assertAlmostEqual(probits.calculate_thermal_dose(heat_flux, exposure_time), 256.4963920)

    def test_calculate_zero_thermal_dose(self):
        heat_flux = 5  # W/m^2
        exposure_time = 30  # seconds
        self.assertAlmostEqual(probits.calculate_thermal_dose(0, exposure_time), 0)
        self.assertAlmostEqual(probits.calculate_thermal_dose(heat_flux, 0), 0)

    def test_thermal_eisenberg(self):
        # Hand-calculation comparison
        thermal_dose = 1e7  # (W/m^2)^(4/3)*s
        self.assertAlmostEqual(probits.thermal_eisenberg(thermal_dose), 2.7823249)

    def test_thermal_tsao_perry(self):
        # Hand-calculation comparison
        thermal_dose = 1e7  # (W/m^2)^(4/3)*s
        self.assertAlmostEqual(probits.thermal_tsao_perry(thermal_dose), 4.8823249)

    def test_thermal_tno(self):
        # Hand-calculation comparison
        thermal_dose = 1e7  # (W/m^2)^(4/3)*s
        self.assertAlmostEqual(probits.thermal_tno(thermal_dose), 4.0323249)

    def test_thermal_lees(self):
        # Hand-calculation comparison
        thermal_dose = 1e7  # (W/m^2)^(4/3)*s
        self.assertAlmostEqual(probits.thermal_lees(thermal_dose), 1.6756475)

    def test_thermal_fatality_prob(self):
        # Heat flux is hand-calculated below to obtrain a 50% probability of fatality with these inputs
        model_ref = 'eise'  # Reference string for Eisenberg thermal probit model
        heat_flux = 15796.170691  # W/m^2
        exposure_time = 60  # seconds
        self.assertAlmostEqual(probits.compute_thermal_fatality_prob(model_ref, heat_flux, exposure_time), 0.5)

    def test_zero_thermal_dose_gives_zero_fatality_prob(self):
        model_ref = 'eise'  # Reference string for Eisenberg thermal probit model
        heat_flux = 0  # W/m^2
        exposure_time = 60  # seconds
        self.assertAlmostEqual(probits.compute_thermal_fatality_prob(model_ref, heat_flux, exposure_time), 0)


class TestOverpressureProbits(unittest.TestCase):
    """
    Tests of overpressure probit calculations
    """
    def test_overpressure_eisenberg(self):
        # Hand-calculation comparison
        peak_overpressure = 1e5  # Pa
        self.assertAlmostEqual(probits.overp_eisenberg(peak_overpressure), 2.4543150)

    def test_overpressure_hse(self):
        # Hand-calculation comparison
        peak_overpressure = 1e5  # Pa
        self.assertAlmostEqual(probits.overp_hse(peak_overpressure), 5.1300000)

    def test_overpressure_tno_head(self):
        # Hand-calculation comparison
        peak_overpressure = 1e5  # Pa
        impulse = 1e4  # Pa*s
        self.assertAlmostEqual(probits.overp_tno_head(peak_overpressure, impulse), 12.2786003)

    def test_overpressure_tno_struct_collapse(self):
        # Hand-calculation comparison
        peak_overpressure = 1e5  # Pa
        impulse = 1e4  # Pa*s
        self.assertAlmostEqual(probits.overp_tno_struct_collapse(peak_overpressure, impulse), 6.4917213)

    def test_overpressure_fatality_prob(self):
        # Peak overpressure is hand-calculated below to obtrain a 50% probability of fatality with these inputs
        model_ref = 'leis'  # Reference string for Eisenburg Lung Hemmorrhage overpressure probit model
        peak_overpressure = 144542.867547  # Pa
        self.assertAlmostEqual(probits.compute_overpressure_fatality_prob(model_ref, peak_overpressure), 0.5)

    def test_zero_overpressure_gives_zero_fatality_prob(self):
        model_ref = 'leis'  # Reference string for Eisenburg Lung Hemmorrhage overpressure probit model
        peak_overpressure = 0
        self.assertAlmostEqual(probits.compute_overpressure_fatality_prob(model_ref, peak_overpressure), 0)

if __name__ == "__main__":
    unittest.main()



"""
Copyright 2015-2023 National Technology & Engineering Solutions of Sandia, LLC (NTESS).
Under the terms of Contract DE-NA0003525 with NTESS, the U.S. Government retains certain rights in this software.

You should have received a copy of the GNU General Public License along with HyRAM+.
If not, see https://www.gnu.org/licenses/.
"""

import copy

import numpy as np
from scipy import optimize

from ._comps import Orifice


class NotionalNozzle:
    def __init__(self, fluid_orifice, orifice, low_P_fluid):
        '''Notional nozzle class'''
        self.fluid_orifice, self.orifice, self.low_P_fluid = fluid_orifice, orifice, low_P_fluid
    
    def calculate(self, T = 'solve_energy', conserve_momentum = True):
        '''
        Calculates the properties after the notional nozzle using one of the following models:
        
        Parameters
        ----------
        T: string ('solve_energy', or 'Tthroat' - if conserve_momentum is false) or value
            how to calculate or specify temperature at end of notional nozzle
        conserve_momentum: boolean
            whether to use conservation of momentum equation
        Literature references for models of different combinations are:
            YuceilOtugen - conserve_momentum = True, T = solve_energy
            EwanMoodie - conserve_momentum = False, T = Tthroat
            Birch - conserve_momentum = False, T = T0
            Birch2 - conserve_momentum = True, T = T0
            Molkov - conserve_momentum = False, T = solve_energy
        
        Returns
        -------
        tuple of (fluid object, orifice object), all at exit of notional nozzle
        '''
        throat = self.fluid_orifice
        if conserve_momentum:
            #YuceilOtugen, Birch2
            v = throat.v + (throat.P - self.low_P_fluid.P)/(throat.v*throat.rho*self.orifice.Cd)
            if T == 'solve_energy':
                #YuceilOtugen
                throat_enthalpy = throat.therm.get_property('H0', P=throat.P, D=throat.rho, v=throat.v)
                rho = optimize.brentq(lambda rho: throat_enthalpy - throat.therm.get_property('H0', P=self.low_P_fluid.P, D=rho, v=v), 
                                      throat.rho, throat.therm.get_property('D', T=self.low_P_fluid.T, P=self.low_P_fluid.P))
            elif np.isreal(T):
                #Birch2 - T specified as T0
                rho = throat.therm.get_property('D', T=T, P=self.low_P_fluid.P)
            else:
                raise NotImplementedError('Notional nozzle model not defined properly, ' + 
                                          "nn_T must be specified temperature or 'solve_energy'.")
        else:
            #EwanMoodie, Birch, Molkov -- assume sonic at notional nozzle location
            if np.isreal(T):
                # Birch: T specified as T0
                v = throat.therm.get_property('A', T=T, P=self.low_P_fluid.P)
                rho = throat.therm.get_property('D', T=T, P=self.low_P_fluid.P)
            elif T == 'Tthroat':
                #EwanMoodie - T is Throat temperature
                v = throat.therm.get_property('A', T=throat.T, P=self.low_P_fluid.P)
                rho = throat.therm.get_property('D', T=throat.T, P=self.low_P_fluid.P)
            elif T == 'solve_energy':
                #Molkov
                def err(rho):
                    T = throat.therm.get_property('T', P=self.low_P_fluid.P, rho=rho)
                    v = throat.therm.get_property('A', T=T, P=self.low_P_fluid.P)
                    return (throat.therm.get_property('H0', P=throat.P, D=throat.rho, v=throat.v) -
                            throat.therm.get_property('H0', P=self.low_P_fluid.P, D=rho, v=v))
                rho = optimize.newton(err, throat.therm.get_property('D', T=self.low_P_fluid.T, P=self.low_P_fluid.P))
                T = throat.therm.get_property('T', P=self.low_P_fluid.P, D=rho)
                v = throat.therm.get_property('A', T=T, P=self.low_P_fluid.P)
            else:
                raise NotImplementedError('Notional nozzle model not defined properly, ' + 
                                          "nn_T must be specified temperature or 'solve_energy'")
        fluid = copy.copy(throat)
        fluid.update(rho=rho, P=self.low_P_fluid.P, v=v)
        # conserve mass to solve for effective diameter:
        orifice = Orifice(np.sqrt(self.orifice.mdot(throat)/(fluid.rho*fluid.v)*4/np.pi))
        return fluid, orifice    



"""
Copyright 2015-2023 National Technology & Engineering Solutions of Sandia, LLC (NTESS).
Under the terms of Contract DE-NA0003525 with NTESS, the U.S. Government retains certain rights in this software.

You should have received a copy of the GNU General Public License along with HyRAM+.
If not, see https://www.gnu.org/licenses/.
"""

import unittest

from hyram.qra import consequence


class TestThermalFatalitiesCalc(unittest.TestCase):
    """
    Tests of calculation of thermal fatality probabilities
    """
    def test_calc_thermal_fatality_probabilities(self):
        consequence_type = 'thermal'
        num_leak_sizes = 5
        total_occupants = 2
        physical_responses = {'qrads': [1000, 2000, 3000, 4000, 5000,
                                        6000, 7000, 8000, 9000, 10000]}  # W/m2
        consequence_modeling_decisions = {'probit_thermal_id': 'eise',  # Eisenberg thermal probit
                                          'exposure_time': 60}  # seconds
        thermal_fatality_probs_per_leak = consequence.calculate_event_consequence(consequence_type=consequence_type,
                                                                                  num_leak_sizes=num_leak_sizes,
                                                                                  total_occupants=total_occupants,
                                                                                  physical_responses=physical_responses,
                                                                                  consequence_modeling_decisions=consequence_modeling_decisions)
        self.assertEqual(len(thermal_fatality_probs_per_leak),
                         num_leak_sizes)


class TestOverpressureFatalitiesCalc(unittest.TestCase):
    """
    Tests of calculation of overpressure fatality probabilities
    """
    def test_calc_overpressure_fatality_probabilities(self):
        consequence_type = 'overp'
        num_leak_sizes = 5
        total_occupants = 2
        physical_responses = {'overpressures': [1000, 2000, 3000, 4000, 5000,
                                                6000, 7000, 8000, 9000, 10000], # Pa
                              'impulses': [1100, 2100, 3100, 4100, 5100,
                                           6100, 7100, 8100, 9100, 11000]}  # Pa*s
        consequence_modeling_decisions = {'probit_overp_id': 'head'}  # TNO Heat Impact overpressure probit
        overpressure_fatality_probs_per_leak = consequence.calculate_event_consequence(consequence_type=consequence_type,
                                                                                       num_leak_sizes=num_leak_sizes,
                                                                                       total_occupants=total_occupants,
                                                                                       physical_responses=physical_responses,
                                                                                       consequence_modeling_decisions=consequence_modeling_decisions)
        self.assertEqual(len(overpressure_fatality_probs_per_leak),
                         num_leak_sizes)


if __name__ == "__main__":
    unittest.main()



"""
Copyright 2015-2023 National Technology & Engineering Solutions of Sandia, LLC (NTESS).
Under the terms of Contract DE-NA0003525 with NTESS, the U.S. Government retains certain rights in this software.

You should have received a copy of the GNU General Public License along with HyRAM+.
If not, see https://www.gnu.org/licenses/.
"""

import os
import unittest

import numpy as np

from hyram.qra import effects
import hyram.phys.api as phys_api
from hyram.phys._jet import DevelopingFlow
from hyram.phys._therm import Combustion
from hyram.phys import Orifice
from hyram.utilities import misc_utils


class TestThermalEffects(unittest.TestCase):
    """
    Test calculation and plotting of thermal effects
    """
    def setUp(self):
        self.amb_fluid = phys_api.create_fluid('AIR',
                                               temp=288,  # K
                                               pres=101325)  # Pa
        self.rel_fluid = phys_api.create_fluid('H2',
                                               temp=288,  # K
                                               pres=35e6,  # Pa
                                               phase='none')
        self.rel_angle = 0  # radians
        self.site_length = 20  # m
        self.site_width = 12  # m
        leak_diams = [0.001, 0.003]  # m
        self.orifices = [Orifice(leak_diam) for leak_diam in leak_diams]
        self.rel_humid = 0.89
        self.not_nozzle_model = 'yuce'
        cons_momentum, notional_noz_t = misc_utils.convert_nozzle_model_to_params(self.not_nozzle_model, self.rel_fluid)
        self.developing_flows = [DevelopingFlow(self.rel_fluid, o, self.amb_fluid,
                                                theta0 = self.rel_angle,
                                                nn_conserve_momentum=cons_momentum, nn_T=notional_noz_t,
                                                verbose=False) for o in self.orifices]
        self.chem = Combustion(self.rel_fluid)
        self.locations = [(5, 0, 1), (6, 1, 2), (7, 0, 2)]  # m
        self.create_plots = True
        self.output_dir = misc_utils.get_temp_folder()
        self.verbose=False

    def test_calc_thermal_effects(self):
        flux_dict = effects.calc_thermal_effects(self.amb_fluid,
                                                 self.rel_fluid,
                                                 self.rel_angle,
                                                 self.site_length,
                                                 self.site_width,
                                                 self.orifices,
                                                 self.rel_humid,
                                                 self.not_nozzle_model,
                                                 self.locations,
                                                 self.developing_flows,
                                                 self.chem,
                                                 self.create_plots,
                                                 self.output_dir,
                                                 self.verbose)
        self.assertGreater(max(flux_dict['fluxes']), 0)
        self.assertGreater(len(flux_dict['all_pos_files']), 0)

        # ensure that specifying a developing flow performs the same as not
        flux_dict_no_develop_flow = \
            effects.calc_thermal_effects(amb_fluid=self.amb_fluid,
                                         rel_fluid=self.rel_fluid,
                                         rel_angle=self.rel_angle,
                                         site_length=self.site_length,
                                         site_width=self.site_width,
                                         orifices=self.orifices,
                                         rel_humid=self.rel_humid,
                                         not_nozzle_model=self.not_nozzle_model,
                                         locations=self.locations,
                                         developing_flows=None,
                                         chem=None,
                                         create_plots=self.create_plots,
                                         output_dir=self.output_dir,
                                         verbose=self.verbose)
        self.assertListEqual(flux_dict['fluxes'].tolist(), flux_dict_no_develop_flow['fluxes'].tolist())

    def test_zero_occupants(self):
        locations = []
        flux_dict = effects.calc_thermal_effects(self.amb_fluid,
                                                 self.rel_fluid,
                                                 self.rel_angle,
                                                 self.site_length,
                                                 self.site_width,
                                                 self.orifices,
                                                 self.rel_humid,
                                                 self.not_nozzle_model,
                                                 locations,
                                                 self.developing_flows,
                                                 self.chem,
                                                 self.create_plots,
                                                 self.output_dir,
                                                 self.verbose)
        self.assertEqual(len(flux_dict['fluxes']), 0)


class TestOverpressureEffects(unittest.TestCase):
    """
    Test calculation and plotting of overpressure effects
    """
    def setUp(self):
        leak_diams = [0.001, 0.003]  # m
        self.orifices = [Orifice(leak_diam) for leak_diam in leak_diams]
        self.notional_nozzle_model = 'yuce'
        self.release_fluid = phys_api.create_fluid('H2',
                                                   temp=288,  # K
                                                   pres=35e6,  # Pa
                                                   phase='none')
        self.ambient_fluid = phys_api.create_fluid('AIR',
                                                   temp=288,  # K
                                                   pres=101325)  # Pa
        self.release_angle = 0  # radians
        cons_momentum, notional_noz_t = misc_utils.convert_nozzle_model_to_params(self.notional_nozzle_model, self.release_fluid)
        self.developing_flows = [DevelopingFlow(self.release_fluid, o, self.ambient_fluid,
                                                theta0 = self.release_angle,
                                                nn_conserve_momentum=cons_momentum, nn_T=notional_noz_t,
                                                verbose=False) for o in self.orifices]
        self.overp_method = 'bst'
        self.locations = [(5, 0, 1), (6, 1, 2), (7, 0, 2)]  # m
        self.site_length = 20  # m
        self.site_width = 12  # m
        self.BST_mach_flame_speed = 0.35
        self.TNT_equivalence_factor = 0.03
        self.create_plots = True
        self.output_dir = None
        self.verbose = False

    def test_calc_overp_effects_TNT(self):
        overp_dict = effects.calc_overp_effects(self.orifices,
                                                self.notional_nozzle_model,
                                                self.release_fluid,
                                                self.ambient_fluid,
                                                self.release_angle,
                                                self.locations,
                                                self.site_length,
                                                self.site_width,
                                                'tnt',
                                                self.BST_mach_flame_speed,
                                                self.TNT_equivalence_factor,
                                                self.developing_flows,
                                                self.create_plots,
                                                self.output_dir,
                                                self.verbose)
        self.assertGreater(max(overp_dict['overpressures']), 0)
        self.assertGreater(max(overp_dict['impulses']), 0)
        self.assertGreater(len(overp_dict['all_pos_overp_files']), 0)
        self.assertGreater(len(overp_dict['all_pos_impulse_files']), 0)

        # ensure that specifying a developing flow performs the same as not
        overp_dict_no_develop_flow = \
            effects.calc_overp_effects(orifices=self.orifices,
                                       notional_nozzle_model=self.notional_nozzle_model,
                                       release_fluid=self.release_fluid,
                                       ambient_fluid=self.ambient_fluid,
                                       release_angle=self.release_angle,
                                       locations=self.locations,
                                       site_length=self.site_length,
                                       site_width=self.site_width,
                                       overp_method='tnt',
                                       BST_mach_flame_speed=self.BST_mach_flame_speed,
                                       TNT_equivalence_factor=self.TNT_equivalence_factor,
                                       developing_flows=None,
                                       create_plots=self.create_plots,
                                       output_dir=self.output_dir,
                                       verbose=self.verbose)
        self.assertListEqual(overp_dict['overpressures'].tolist(),
                             overp_dict_no_develop_flow['overpressures'].tolist())


    def test_calc_overp_effects_BST(self):
        overp_dict = effects.calc_overp_effects(self.orifices,
                                                self.notional_nozzle_model,
                                                self.release_fluid,
                                                self.ambient_fluid,
                                                self.release_angle,
                                                self.locations,
                                                self.site_length,
                                                self.site_width,
                                                'bst',
                                                self.BST_mach_flame_speed,
                                                self.TNT_equivalence_factor,
                                                self.developing_flows,
                                                self.create_plots,
                                                self.output_dir,
                                                self.verbose)
        self.assertGreater(max(overp_dict['overpressures']), 0)
        self.assertGreater(max(overp_dict['impulses']), 0)
        self.assertGreater(len(overp_dict['all_pos_overp_files']), 0)
        self.assertGreater(len(overp_dict['all_pos_impulse_files']), 0)

    def test_calc_overp_effects_Bauwens(self):
        overp_dict = effects.calc_overp_effects(self.orifices,
                                                self.notional_nozzle_model,
                                                self.release_fluid,
                                                self.ambient_fluid,
                                                self.release_angle,
                                                self.locations,
                                                self.site_length,
                                                self.site_width,
                                                'bauwens',
                                                self.BST_mach_flame_speed,
                                                self.TNT_equivalence_factor,
                                                self.developing_flows,
                                                self.create_plots,
                                                self.output_dir,
                                                self.verbose)
        self.assertGreater(max(overp_dict['overpressures']), 0)
        self.assertTrue(np.all(np.isnan(overp_dict['impulses'])))
        self.assertGreater(len(overp_dict['all_pos_overp_files']), 0)
        self.assertEqual(len(overp_dict['all_pos_impulse_files']), 0)

    def test_zero_occupants(self):
        locations = []
        overp_dict = effects.calc_overp_effects(self.orifices,
                                                self.notional_nozzle_model,
                                                self.release_fluid,
                                                self.ambient_fluid,
                                                self.release_angle,
                                                locations,
                                                self.site_length,
                                                self.site_width,
                                                self.overp_method,
                                                self.BST_mach_flame_speed,
                                                self.TNT_equivalence_factor,
                                                self.developing_flows,
                                                self.create_plots,
                                                self.output_dir,
                                                self.verbose)
        self.assertEqual(len(overp_dict['overpressures']), 0)
        self.assertEqual(len(overp_dict['impulses']), 0)


class TestEffectPlots(unittest.TestCase):
    """
    Test plotting of thermal and overpressure effects
    """
    def test_plot_effect_positions(self):
        test_effects = [0.06, 0.01]
        effect_label = 'Test Effect [units]'
        output_dir = misc_utils.get_temp_folder()
        filename = 'test_effect_positions.png'
        filepathname = os.path.join(output_dir, filename)
        title = 'Test Effect Positions Plot'
        x_locations = [1, 5]
        z_locations = [1, 2]
        length = 20
        width = 12
        effects.plot_effect_positions(test_effects,
                                      effect_label,
                                      filepathname,
                                      title,
                                      x_locations,
                                      z_locations,
                                      length,
                                      width)
        self.assertTrue(os.path.isfile(filepathname))


if __name__ == "__main__":
    unittest.main()



"""
Copyright 2015-2023 National Technology & Engineering Solutions of Sandia, LLC (NTESS).
Under the terms of Contract DE-NA0003525 with NTESS, the U.S. Government retains certain rights in this software.

You should have received a copy of the GNU General Public License along with HyRAM+.
If not, see https://www.gnu.org/licenses/.

"""
import os
import unittest

import numpy as np
import matplotlib.pyplot as plt

from hyram.phys import Orifice, Fluid, Jet, Flame
from hyram.phys._plots import plot_contour
from hyram.utilities import misc_utils

"""
NOTE: if running from IDE like pycharm, make sure cwd is hyram/ and not hyram/tests.
"""
VERBOSE = False


class TestContourPlots(unittest.TestCase):
    def setUp(self):
        t_amb = 300  # K
        orifice_diameter = 0.0254  # m
        dischage_coefficent = 1
        ambient_pressure = 101325  # Pa
        p_rel = 60e5  # Pa

        orifice = Orifice(orifice_diameter, dischage_coefficent)
        amb_fluid = Fluid(species='air', P=ambient_pressure, T=t_amb)
        rel_fluid = Fluid(species='hydrogen', T=t_amb, P=p_rel)

        nozzle_cons_m, nozzle_t_param = misc_utils.convert_nozzle_model_to_params('yuce', rel_fluid)

        self.jet_object_horizontal = Jet(rel_fluid, orifice, amb_fluid,
                                        nn_conserve_momentum=nozzle_cons_m,
                                        nn_T=nozzle_t_param,
                                        verbose=VERBOSE)

        self.jet_object_vertical = Jet(rel_fluid, orifice, amb_fluid,
                                        theta0=np.pi/2,
                                        nn_conserve_momentum=nozzle_cons_m,
                                        nn_T=nozzle_t_param,
                                        verbose=VERBOSE)

        self.flame_object = Flame(rel_fluid, orifice, amb_fluid)

    def test_auto_limits_horizontal(self):
        fig = plot_contour(data_type="mole", jet_or_flame=self.jet_object_horizontal, vlims=None)
        self.assertTrue(fig is not None)

        filename = os.path.join(misc_utils.get_temp_folder(), 'test_auto_limits_h.png')
        fig.savefig(filename)
        plt.close(fig)

    def test_auto_limits_vertical(self):
        fig = plot_contour(data_type="mole", jet_or_flame=self.jet_object_vertical, vlims=None)
        self.assertTrue(fig is not None)

        filename = os.path.join(misc_utils.get_temp_folder(), 'test_auto_limits_v.png')
        fig.savefig(filename)
        plt.close(fig)

    def test_set_limits_horizontal(self):
        fig = plot_contour(data_type="mole", jet_or_flame=self.jet_object_horizontal, xlims=(0, 20), ylims=(-2, 2), contour_levels=0.1)
        self.assertTrue(fig is not None)

        filename = os.path.join(misc_utils.get_temp_folder(), 'test_set_limits_h.png')
        fig.savefig(filename)
        plt.close(fig)

    def test_set_limits_vertical(self):
        fig = plot_contour(data_type="mole", jet_or_flame=self.jet_object_vertical, xlims=(-2, 2), ylims=(0, 20), contour_levels=0.1)
        self.assertTrue(fig is not None)

        filename = os.path.join(misc_utils.get_temp_folder(), 'test_set_limits_v.png')
        fig.savefig(filename)
        plt.close(fig)

    def test_temp_plot(self):
        fig = plot_contour(data_type="temperature", jet_or_flame=self.flame_object, xlims=(0, 20), ylims=(-2.5, 5))
        self.assertTrue(fig is not None)

        filename = os.path.join(misc_utils.get_temp_folder(), 'test_temp.png')
        fig.savefig(filename)
        plt.close(fig)

    def test_invalid_input(self):
        with self.assertRaises(ValueError):
            plot_contour(data_type="test", jet_or_flame=self.jet_object_horizontal)

    def test_invalid_contour_levels(self):
        with self.assertRaises(ValueError):
            plot_contour(data_type="mole", jet_or_flame=self.jet_object_horizontal, contour_levels=0)
        with self.assertRaises(ValueError):
            plot_contour(data_type="mole", jet_or_flame=self.jet_object_horizontal, contour_levels=1)
        with self.assertRaises(ValueError):
            plot_contour(data_type="mole", jet_or_flame=self.jet_object_horizontal, contour_levels=[0,1])


if __name__ == "__main__":
    unittest.main()



"""
Copyright 2015-2023 National Technology & Engineering Solutions of Sandia, LLC (NTESS).
Under the terms of Contract DE-NA0003525 with NTESS, the U.S. Government retains certain rights in this software.

You should have received a copy of the GNU General Public License along with HyRAM+.
If not, see https://www.gnu.org/licenses/.
"""

import unittest

import hyram.phys.api as phys_api
from hyram.phys import Orifice, Flame
from hyram.phys._flame import calc_transmissivity


VERBOSE = False


class TestAtmosphericTransmissivity(unittest.TestCase):
    """
    Test calculation of atmospheric transmissivity
    """
    def test_against_values_from_paper(self):
        path_lengths = [
            10, 100, 1000,
            10, 100, 1000,
            10, 100, 1000,
            10, 100, 1000
        ]  # m
        temperatures = [
            253, 253, 253,
            273, 273, 273,
            293, 293, 293,
            303, 303, 303
        ]  # K
        relative_humidities = [
            0.5, 0.5, 0.5,
            0.5, 0.5, 0.5,
            0.5, 0.5, 0.5,
            0.5, 0.5, 0.5
        ]
        calc_taus = []
        for path_length, amb_temp, rel_humid in zip(path_lengths,
                                                    temperatures,
                                                    relative_humidities):
            tau = calc_transmissivity(path_length, amb_temp, rel_humid)
            calc_taus.append(tau)
        taus_from_paper = [
            0.96, 0.86, 0.72,
            0.91, 0.78, 0.61,
            0.86, 0.71, 0.51,
            0.84, 0.67, 0.47
        ]
        for calc_tau, paper_tau in zip(calc_taus, taus_from_paper):
            tau_diff_pct = (calc_tau - paper_tau) / paper_tau * 100
            self.assertLessEqual(tau_diff_pct, 0.6)


class TestFlameObject(unittest.TestCase):
    """
    Tests of Flame class
    """
    def setUp(self):
        release_fluid = phys_api.create_fluid('H2',
                                              temp=288,  # K
                                              pres=35e6,  # Pa
                                              phase='none')
        ambient_fluid = phys_api.create_fluid('AIR',
                                              temp=288,  # K
                                              pres=101325)  # Pa
        leak_diam = 0.003  # m
        orifice = Orifice(leak_diam)
        flame = Flame(release_fluid,
                      orifice,
                      ambient_fluid,
                      verbose=VERBOSE)
        self.flame = flame

    def test_zero_occupants_positional_flux(self):
        locations = []
        rel_humid = 0.5
        fluxes = self.flame.generate_positional_flux(locations, rel_humid)
        self.assertEqual(len(fluxes), 0)


if __name__ == "__main__":
    unittest.main()



"""
Copyright 2015-2023 National Technology & Engineering Solutions of Sandia, LLC (NTESS).
Under the terms of Contract DE-NA0003525 with NTESS, the U.S. Government retains certain rights in this software.

You should have received a copy of the GNU General Public License along with HyRAM+.
If not, see https://www.gnu.org/licenses/.
"""

import warnings

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import scipy.constants as const
from matplotlib.collections import LineCollection
from scipy import interpolate

from ._fuel_props import FuelProperties
from ._comps import Fluid, Orifice
from ._layer import LayeringJet
from ._therm import Combustion
from ..utilities import misc_utils
from ..utilities.custom_warnings import PhysicsWarning



########################################################################
# TODO: untested for alternate fuels
########################################################################
class IndoorRelease:
    '''
    Class used to calculate physics of an indoor release
    '''
    def __init__(self, source, orifice, ambient, enclosure, tmax=None,
                 heat_flux=0, nmax=1000, m_empty=1e-6, p_empty_percent=1e-2,
                 release_area=None, theta0=0, x0=0, y0=0,
                 nn_conserve_momentum=True, nn_T='solve_energy',
                 lam=1.16, X_lean=None, X_rich=None, tol=1e-5,
                 max_steps=1e5, steady=False, nsteady=5,
                 verbose=False):
        '''
        Initialization for an indoor release
        
        Parameters
        ----------
        source : class
            source for fluid (must contain empty function)
        orifice : class
            orifice through which source is flowing
        ambient : class
            fluid initially contained in enclosure
        enclosure : class
            enclosure into which release is occurring
        tmax : float (optional)
            total time for simulation (s), unless blowdown not complete
        heat_flux: float (optional)
            Heat flow (W) into source tank.  Assumed to be 0 (adiabatic)
        nmax: int (optional)
            maximum number of iterations for blowdown integration
        m_empty: float (optional)
            mass when source considered empty (kg)
        p_empty_percent: float (optional)
            percent of ambient pressure when source considered empty
        release_area : float or None (optional)
            secondary containment area for release (m^2)
        theta0 : float (optional)
            angle of release (rad, 0 is horizontal, pi/2 is vertical)
            Default is 0
        x0 : float (optional)
            x-coordinate of release (m), default is 0
        y0 : float (optional)
            y-coordinate of release (m), default is 0
        lam : float (optional)
            lambda value, default is 1.16
        X_lean : float (optional)
            molar lower flammability limit,
            default is None: will use default LFL for fuel
        X_rich : float (optional)
            molar upper flammability limit,
            default is None: will use default UFL for fuel
        tol : float (optional)
            Tolerance for h2_jet integration, default is 1e-5
        max_steps : integer (optional)
            Maximum steps along the S axis for h2_jet integration
            Default is 1e5
        steady : Boolean (optional)
            Option for a blowdown release or a steady-flowrate release
            Default is False, for a blowdown
        nsteady: integer (optional)
            number of time divisions to get to steady state
        '''
        if X_lean is None:
            fuel_props = FuelProperties(source.fluid.species)
            X_lean = fuel_props.LFL
        if X_rich is None:
            fuel_props = FuelProperties(source.fluid.species)
            X_rich = fuel_props.UFL

        if verbose:
            params = locals()
            print("=== BEGINNING INDOOR RELEASE ===")
            print("PARAMETERS")
            print(misc_utils.params_as_str(params))
        
        # Calculate time steps and mass flow history
        if steady:
            # Single jet/plume for steady release, so single time step
            if tmax is None:
                tmax = 30 # this is an arbitrary number - should be long enough to ensure steady-state has been reached
            ts = np.linspace(0, tmax, nsteady) 
            steady_mdot = orifice.mdot(orifice.flow(source.fluid, ambient.P))
            mdots = np.ones(len(ts)) * steady_mdot
            gas_list = [source.fluid for i in range(len(ts))]
        else:
            # Different jet/plume at each time step for blowdown
            mdots, gas_list, ts, _ = source.empty(orifice, ambient.P,
                                                  heat_flux, nmax, m_empty, p_empty_percent)
            gas_list = gas_list.copy()
            if tmax is not None:
                if tmax > ts[-1]:
                    ts = np.append(ts, tmax)
                    gas_list.append(Fluid(species = source.fluid.species, T = ambient.T, P = ambient.P))
                    mdots = np.append(mdots, 1e-10)
                else:
                    i = np.argmax(np.array(ts) > tmax)+1
                    ts = ts[:i]
                    gas_list = gas_list[:i]
                    mdots = mdots[:i]
        # Source fluid at ambient conditions
        gas = Fluid(species = source.fluid.species, T = ambient.T, P = ambient.P)
        self.comb = Combustion(gas)
        self.enclosure = enclosure
        
        if release_area is not None:
            if release_area > orifice.A:
                # switch gas to ambient pressure and change orifice size for release
                gas_list = len(gas_list)*[gas]
                orifice = Orifice(np.sqrt(release_area*4/const.pi))
            else:
                warnings.warn('Secondary containment release area must be bigger than orifice area - assuming single orifice.',
                              category=PhysicsWarning)
        jets = []
        LIM = enclosure.Xwall + enclosure.H # Limit of maximum distance jet can extend
        Ymin = X_lean*gas.therm.MW/(X_lean*gas.therm.MW + (1-X_lean)*ambient.therm.MW)
        for g, mdot in zip(gas_list, mdots):
            jets.append(LayeringJet(g, orifice, ambient, theta0 = theta0, y0 = y0,
                                    nn_conserve_momentum=nn_conserve_momentum, nn_T=nn_T,
                                    x0 = x0, lam = lam, mdot = mdot, Smax = LIM, Ymin = Ymin,
                                    max_steps = max_steps, tol = tol, suppressWarnings = True, verbose = verbose))
            jets[-1].Q_jet = mdot/gas.rho # needed for layer model
        # Reshape jet if needed
        [jet.reshape(enclosure, showPlot=False) for jet in jets]

        # Initial 'layer' is between vent and ceiling
        vol_layer = (enclosure.H - enclosure.ceiling_vent.H) * enclosure.A

        # Set initial volume and concentration (0) for layer model
        Vol_conc0 = np.array([vol_layer, 0])

        # Initialize solutions
        x_layer, H_layer, m_jet, m_layer, dP_layer, dP_tot = 6*[np.array([0])]
        Vol_layer, t_layer = 2*[np.array([0])]
        
        jet_mass_last = jets[0].m_flammable(X_lean = X_lean, X_rich = X_rich, Hmax = enclosure.H)
        # Run layer model at each time step for each plume
        for i in range(0, len(jets)-1):
            # Compute volumetric distribution of gas along the ceiling
            t, vol_layer, c = jets[i].layer_accumulation([ts[i], ts[i+1]], Vol_conc0, enclosure)
            # Ensure volume and mole fraction values are realistic
            if np.any(vol_layer < 0): raise ValueError('Layer volume has returned a negative value')
            if np.any(vol_layer > enclosure.V): raise ValueError('Layer volume has exceeded enclosure volume')
            if np.any(c < 0): raise ValueError('Layer concentration has returned a negative value')
            if np.any(c > 1): raise ValueError('Layer concentration has exceeded 100%')
            
            # Calculate flammable mass in layer
            MW_layer = c*gas.therm.MW + (1 - c)*ambient.therm.MW
            rho_layer = ambient.P/(ambient.T*const.R/MW_layer)
            Y_layer = c * gas.therm.MW / MW_layer
            layer_mass = vol_layer*rho_layer*Y_layer*(c >= X_lean)*(c <= X_rich)
            
            # Calculate height of layer
            layer_height = vol_layer / enclosure.A

            # Calculate flammable mass in jet
            jet_mass = jets[i+1].m_flammable(X_lean = X_lean, X_rich = X_rich, Hmax = enclosure.H - layer_height[-1])
            jet_mass_array = np.interp(t, [ts[i], ts[i+1]], [jet_mass_last, jet_mass])
            jet_mass_last = jet_mass

            # Calculate total overpressure
            dP_total = self.comb.dP_expansion(enclosure, jet_mass_array + layer_mass)

            # Calculate overpressure in layer
            dP_lay = self.comb.dP_expansion(enclosure, layer_mass)

            # Assign outputs for this plume-timestep
            x_layer = np.append(x_layer, c[1:])
            Vol_layer = np.append(Vol_layer, vol_layer[1:])
            m_layer = np.append(m_layer, layer_mass[1:])
            H_layer = np.append(H_layer, layer_height[1:])
            m_jet = np.append(m_jet, jet_mass_array[1:])
            dP_tot = np.append(dP_tot, dP_total[1:])
            dP_layer = np.append(dP_layer, dP_lay[1:])
            t_layer = np.append(t_layer, t[1:])

            # Update initial volume and concentration for layer model
            Vol_conc0 = np.array([vol_layer[-1], c[-1]])
        
        if verbose:
            print('done')
        
        # Assign outputs
        self.ts, self.mdots = ts, mdots
        self.t_layer = t_layer
        self.x_layer, self.H_layer = x_layer, H_layer
        self.m_jet, self.m_layer = m_jet, m_layer
        self.dP_layer, self.dP_tot = dP_layer, dP_tot
        self.Vol_layer, self.plumes = Vol_layer, jets

    def plot_trajectories(self):
        fig, ax = plt.subplots()
        lines = [list(zip(pl.x, pl.y)) for pl in self.plumes]
        line_segments = LineCollection(lines, norm = mpl.colors.LogNorm())
        line_segments.set_array(np.array(self.ts))
        ax.add_collection(line_segments)
        ax.autoscale()
        axcb = fig.colorbar(line_segments)
        axcb.set_label('Time [s]')
        ax.set_xlabel('x [m]')
        ax.set_ylabel('y [m]')
        ax.set_title('Release Path Trajectories Over Time')
        return fig

    def plot_mass_flows(self):
        fig, ax = plt.subplots()
        ax.semilogy(self.ts, self.mdots)
        ax.set_xlabel('Time [s]')
        ax.set_ylabel('Fuel Mass Flow Rate [kg/s]')
        return fig
    
    def plot_layer(self):
        fig, ax = plt.subplots(2, 1, sharex = True)
        l1 = ax[0].plot(self.t_layer, np.array(self.x_layer)*100,
                     label = 'Mole Fraction Fuel')
        ax[0].set_ylabel('%Fuel in Layer\n(Molar or Volume)')
        i = np.argmin(np.abs(self.t_layer - (np.max(self.t_layer)
                                        - np.min(self.t_layer)) / 2))
        l2 = ax[1].plot(self.t_layer, self.H_layer,
                      'g', label="Height of Layer")
        ax[1].set_ylabel('Layer Thickness  [m]\n(From Ceiling)')
        ax[1].set_xlabel('Time [s]')
        return fig

    def plot_mass(self):
        fig, ax = plt.subplots()
        ax.plot(self.t_layer, self.m_jet, label = 'Plume')
        ax.plot(self.t_layer, self.m_layer, label = 'Layer')
        ax.plot(self.t_layer, np.array(self.m_jet)+np.array(self.m_layer),
                 label='Combined')
        ax.set_xlabel('Time [s]')
        ax.set_ylabel('Flammable Fuel Mass [kg]')
        ax.legend(ncol=3, loc='lower center', bbox_to_anchor=(0.5, 1),
                  fancybox=True)
        return fig

    def plot_overpressure(self, data=None, limit=None):
        fig, ax = plt.subplots()
        ax.plot(self.t_layer, np.array(self.dP_layer)/1000, label='Layer')
        ax.plot(self.t_layer, np.array(self.dP_tot)/1000, label='Combined')
        if data is not None:
            i = 0
            cs = ['b', 'g', 'r', 'c', 'm', 'y', 'k']
            for d in data:
                ax.plot([d[0]], [d[1]], 'o', color=cs[i])
                i += 1
        if limit is not None:
            for l in limit:
                ax.axhline(l, color='k', dashes=(1, 1))
        ax.set_xlabel('Ignition Delay Time [s]')
        ax.set_ylabel('Overpressure [kPa]')
        ax.legend(ncol=2, loc='lower center', bbox_to_anchor=(0.5, 1),
                  fancybox=True)
        return fig

    def pressure(self, t):
        '''
        Returns pressure at time t (or times ts)
        
        Parameters
        -----------
        t : ndarray
           time(s) at which to return the pressure (s)
        
        Returns
        -------
        dP : ndarray
           overpressure(s) at time t (Pa)
        '''
        dp = interpolate.interp1d(self.t_layer, self.dP_tot, bounds_error = False)
        return dp(t)

    def layer_depth(self, t):
        '''
        Returns depth of layer at time t (or times ts)
        
        Parameters
        -----------
        t : ndarray
           time(s) at which to return the depth (s)
        
        Returns
        -------
        ld : ndarray
           layer height(s) at time t (m)
        '''
        ld = interpolate.interp1d(self.t_layer, self.H_layer, bounds_error = False)
        return ld(t)

    def concentration(self, t):
        '''
        Returns layer concentration at time t (or times ts)
        
        Parameters
        -----------
        t : ndarray
           time(s) at which to return the concentration(s)
        
        Returns
        -------
        lc : ndarray
           concentrations(s) at time t (%)
        '''
        lc = interpolate.interp1d(self.t_layer, self.x_layer, bounds_error = False)
        return 100 * lc(t)

    def max_p_t(self):
        '''
        Returns the maximum overpressure and time at it occurs
        
        Returns
        -------
        p_t : tuple
           maximum overpressure (Pa) and time when it occurs (s)
        '''
        imax = np.argmax(self.dP_tot)
        return self.dP_tot[imax], self.t_layer[imax]



"""
Copyright 2015-2023 National Technology & Engineering Solutions of Sandia, LLC (NTESS).
Under the terms of Contract DE-NA0003525 with NTESS, the U.S. Government retains certain rights in this software.

You should have received a copy of the GNU General Public License along with HyRAM+.
If not, see https://www.gnu.org/licenses/.
"""

import unittest

from hyram.phys import Fluid, Jet, Orifice


VERBOSE = False


class TestJetObject(unittest.TestCase):
    """
    Tests of Jet class
    """
    def setUp(self):
        release_fluid = Fluid('H2',
                              T=298,  # K
                              P=35e6)  # Pa
        ambient_fluid = Fluid('AIR',
                              T=298,  # K
                              P=101325)  # Pa
        orifice = Orifice(d=0.003)  # m
        jet = Jet(release_fluid,
                  orifice,
                  ambient_fluid,
                  verbose=VERBOSE)
        self.jet_obj = jet

    def test_get_xy_distances_to_molefrac_horizontal(self):
        #  Default jet object input for theta0 is 0, which is horizontal
        mole_fraction = 0.04
        distances = self.jet_obj.get_xy_distances_to_mole_fractions(mole_fraction)
        # self.assertEqual(distances[mole_fraction]['x']['min'], 0)
        # self.assertGreater(distances[mole_fraction]['x']['max'], 0)
        # self.assertLess(distances[mole_fraction]['y']['min'], 0)
        # self.assertGreater(distances[mole_fraction]['y']['max'], 0)
        self.assertEqual(distances[mole_fraction][0][0], 0)
        self.assertGreater(distances[mole_fraction][0][1], 0)
        self.assertLess(distances[mole_fraction][1][0], 0)
        self.assertGreater(distances[mole_fraction][1][1], 0)

    def test_get_xy_distances_to_molefrac_vertical(self):
        vertical_jet = Jet(self.jet_obj.fluid,
                           self.jet_obj.developing_flow.orifice,
                           self.jet_obj.ambient,
                           theta0=1.570796)  # 90 deg
        mole_fraction = 0.04
        distances = vertical_jet.get_xy_distances_to_mole_fractions(mole_fraction)
        self.assertLess(distances[mole_fraction][0][0], 0)
        self.assertGreater(distances[mole_fraction][0][1], 0)
        self.assertEqual(distances[mole_fraction][1][0], 0)
        self.assertGreater(distances[mole_fraction][1][1], 0)
        # self.assertLess(distances[mole_fraction]['x']['min'], 0)
        # self.assertGreater(distances[mole_fraction]['x']['max'], 0)
        # self.assertEqual(distances[mole_fraction]['y']['min'], 0)
        # self.assertGreater(distances[mole_fraction]['y']['max'], 0)

    def test_get_xy_distances_to_multiple_molefracs(self):
        mole_fractions = [0.04, 0.1]
        distances = self.jet_obj.get_xy_distances_to_mole_fractions(mole_fractions)
        self.assertGreater(list(distances.keys())[1], list(distances.keys())[0])
        # self.assertGreater(distances[mole_fractions[0]]['x']['max'],
        #                    distances[mole_fractions[1]]['x']['max'])
        self.assertGreater(distances[mole_fractions[0]][0][1],
                           distances[mole_fractions[1]][0][1])


if __name__ == "__main__":
    unittest.main()
